import inspect
from functools import lru_cache
from pathlib import Path

from nettacker import version
from nettacker.core.utils.common import generate_random_token, now

CWD = Path.cwd()
PACKAGE_PATH = Path(__file__).parent


@lru_cache(maxsize=128)
def version_info():
    """
    version information of the framework

    Returns:
        an array of version and code name
    """

    return version.__version__, version.__release_name__


class ConfigBase:
    @classmethod
    def as_dict(cls):
        return {attr_name: getattr(cls, attr_name) for attr_name in cls()}

    def __init__(self) -> None:
        self.attributes = sorted(
            (
                attribute[0]
                for attribute in inspect.getmembers(self)
                if not attribute[0].startswith("_") and not inspect.ismethod(attribute[1])
            )
        )
        self.idx = 0

    def __iter__(self):
        yield from self.attributes


# Some sensitive header fields for HTTP requests.
# Please edit this if you don't want your HTTP header to be present in the logs
sensitive_headers = {
    "authorization",
    "proxy-authorization",
    "cookie",
    "set-cookie",
    "x-api-key",
    "x-amz-security-token",
    "x-amz-credential",
    "x-amz-signature",
    "x-session-id",
    "x-csrf-token",
    "x-auth-token",
    "x-user-token",
    "x-id-token",
}


class ApiConfig(ConfigBase):
    """OWASP Nettacker API Default Configuration"""

    api_access_log = str(CWD / ".nettacker/data/nettacker.log")
    api_access_key = generate_random_token(32)
    api_client_whitelisted_ips = []  # disabled - to enable please put an array with list of ips/cidr/ranges
    # [
    #     "127.0.0.1",
    #     "10.0.0.0/24",
    #     "192.168.1.1-192.168.1.255"
    # ],
    api_debug_mode = False
    api_hostname = "0.0.0.0"
    api_port = 5000
    start_api_server = False


class DbConfig(ConfigBase):
    """
    Database Config (could be modified by user)
    For sqlite database:
        fill the name of the DB as sqlite,
        DATABASE as the name of the db user wants
        Set the journal_mode (default="WAL") and
        synchronous_mode (default="NORMAL"). Rest
        of the fields can be left empty
        This is the default database:
        str(CWD / ".nettacker/data/nettacker.db")
    For mysql users:
        fill the ENGINE name of the DB as mysql
        NAME as the name of the database you want to create
        USERNAME, PASSWORD, HOST and the PORT of the MySQL server
        need to be filled respectively (default port is 3306)
    For postgres users:
        fill the Engine name of the DB as postgres
        NAME as the name of the database user wants
        USERNAME, PASSWORD, HOST and the PORT of the Postgres server
        need to be filled respectively (default port is 5432)
        Set ssl_mode to "require" if you need to use encrypted
        databases.
    """

    engine = "sqlite"
    name = str(CWD / ".nettacker/data/nettacker.db")
    host = ""
    port = ""
    username = ""
    password = ""
    ssl_mode = "disable"
    journal_mode = "WAL"
    synchronous_mode = "NORMAL"


class PathConfig:
    """
    home path for the framework (could be modify by user)

    Returns:
        a JSON contain the working, tmp and results path
    """

    data_dir = CWD / ".nettacker/data"
    new_database_file = CWD / ".nettacker/data/nettacker.db"
    old_database_file = CWD / ".data/nettacker.db"
    graph_dir = PACKAGE_PATH / "lib/graph"
    home_dir = CWD
    locale_dir = PACKAGE_PATH / "locale"
    logo_file = PACKAGE_PATH / "logo.txt"
    module_protocols_dir = PACKAGE_PATH / "core/lib"
    modules_dir = PACKAGE_PATH / "modules"
    payloads_dir = PACKAGE_PATH / "lib/payloads"
    release_name_file = PACKAGE_PATH / "release_name.txt"
    results_dir = CWD / ".nettacker/data/results"
    tmp_dir = CWD / ".nettacker/data/tmp"
    web_static_dir = PACKAGE_PATH / "web/static"
    user_agents_file = PACKAGE_PATH / "lib/payloads/User-Agents/web_browsers_user_agents.txt"


class DefaultSettings(ConfigBase):
    """OWASP Nettacker Default Configuration"""

    excluded_modules = None
    excluded_ports = None
    graph_name = "d3_tree_v2_graph"
    language = "en"
    modules_extra_args = None
    parallel_module_scan = 1
    passwords = None
    passwords_list = None
    # Setting to toggle between APSW and SQLAlchemy for sqlite databases.
    use_apsw_for_sqlite = True

    ping_before_scan = False
    ports = None
    profiles = None
    schema = None
    report_path_filename = "{results_path}/results_{date_time}_{random_chars}.html".format(
        results_path=PathConfig.results_dir,
        date_time=now(format="%Y_%m_%d_%H_%M_%S"),
        random_chars=generate_random_token(10),
    )
    retries = 1
    max_retries = 3
    max_submit_query_retry = 100
    retry_delay = 0.1
    scan_ip_range = False
    scan_subdomains = False
    selected_modules = None
    url_base_path = None
    http_header = None
    read_from_file = ""
    set_hardware_usage = "maximum"  # low, normal, high, maximum
    show_all_modules = False
    show_all_profiles = False
    show_help_menu = False
    show_version = False
    skip_service_discovery = False
    socks_proxy = None
    targets = None
    targets_list = None
    thread_per_host = 100
    time_sleep_between_requests = 0.0
    timeout = 3.0
    user_agent = "Nettacker {version_number} {version_code}".format(
        version_number=version_info()[0], version_code=version_info()[1]
    )
    usernames = None
    usernames_list = None
    verbose_event = False
    verbose_mode = False
    scan_compare_id = None
    compare_report_path_filename = ""


class Config:
    api = ApiConfig()
    db = DbConfig()
    path = PathConfig()
    settings = DefaultSettings()



import json

from nettacker.config import Config
from nettacker.core.messages import messages


def escape_for_html_js(json_str: str) -> str:
    """
    This is necessary because some payloads have HTML tags for XSS
    as in waf.yaml, which break the HTML and output no graph. These are unicode escape
    characters for the same
    """
    return json_str.replace("<", "\\u003C").replace(">", "\\u003E").replace("&", "\\u0026")


def start(events):
    """
    generate the d3_tree_v1_graph with events

    Args:
        events: all events

    Returns:
        a graph in HTML
    """

    # define  a normalised_json
    normalisedjson = {"name": "Started attack", "children": {}}
    # get data for normalised_json
    for event in events:
        if event["target"] not in normalisedjson["children"]:
            normalisedjson["children"].update({event["target"]: {}})
            normalisedjson["children"][event["target"]].update({event["module_name"]: []})

        if event["module_name"] not in normalisedjson["children"][event["target"]]:
            normalisedjson["children"][event["target"]].update({event["module_name"]: []})
        normalisedjson["children"][event["target"]][event["module_name"]].append(
            f"target: {event['target']}, module_name: {event['module_name']}, port: "
            f"{event['port']}, event: {event['event']}"
        )
    # define a d3_structure_json
    d3_structure = {"name": "Starting attack", "children": []}
    # get data for normalised_json
    for target in list(normalisedjson["children"].keys()):
        for module_name in list(normalisedjson["children"][target].keys()):
            for description in normalisedjson["children"][target][module_name]:
                children_array = [{"name": module_name, "children": [{"name": description}]}]
                d3_structure["children"].append({"name": target, "children": children_array})

    data = (
        open(Config.path.web_static_dir / "report/d3_tree_v1.html")
        .read()
        .replace("__data_will_locate_here__", escape_for_html_js(json.dumps(d3_structure)))
        .replace("__title_to_replace__", messages("pentest_graphs"))
        .replace("__description_to_replace__", messages("graph_message"))
        .replace("__html_title_to_replace__", messages("nettacker_report"))
    )
    return data



import datetime


def now(format="%Y-%m-%d %H:%M:%S"):
    """
    get now date and time
    Args:
        format: the date and time model, default is "%Y-%m-%d %H:%M:%S"

    Returns:
        the date and time of now
    """
    return datetime.datetime.now().strftime(format)



import copy

import yaml

from nettacker.config import Config


class TemplateLoader:
    def __init__(self, name, inputs=None) -> None:
        self.name = name
        self.inputs = inputs or {}

    @staticmethod
    def parse(module_content, module_inputs):
        if isinstance(module_content, dict):
            for key in copy.deepcopy(module_content):
                if key in module_inputs:
                    if module_inputs[key]:
                        module_content[key] = module_inputs[key]
                elif isinstance(module_content[key], (dict, list)):
                    module_content[key] = TemplateLoader.parse(module_content[key], module_inputs)
        elif isinstance(module_content, list):
            array_index = 0
            for key in copy.deepcopy(module_content):
                module_content[array_index] = TemplateLoader.parse(key, module_inputs)
                array_index += 1

        return module_content

    def open(self):
        module_name_parts = self.name.split("_")
        action = module_name_parts[-1]
        library = "_".join(module_name_parts[:-1])

        with open(Config.path.modules_dir / action / f"{library}.yaml") as yaml_file:
            return yaml_file.read()

    def format(self):
        return self.open().format(**self.inputs)

    def load(self):
        return self.parse(yaml.safe_load(self.format()), self.inputs)



import logging

from paramiko import AutoAddPolicy, SSHClient
from paramiko.auth_strategy import NoneAuth, Password

from nettacker.core.lib.base import BaseEngine, BaseLibrary

logging.getLogger("paramiko.transport").disabled = True


class SshLibrary(BaseLibrary):
    client = SSHClient

    def brute_force(self, *args, **kwargs):
        host = kwargs["host"]
        port = kwargs["port"]
        username = kwargs["username"]
        password = kwargs["password"]

        connection = self.client()
        connection.set_missing_host_key_policy(AutoAddPolicy())
        connection.connect(
            **{
                "hostname": host,
                "port": port,
                "auth_strategy": Password(username=username, password_getter=lambda: password)
                if password
                else NoneAuth(username=username),
            }
        )

        return {
            "host": host,
            "port": port,
            "username": username,
            "password": password,
        }


class SshEngine(BaseEngine):
    library = SshLibrary



import poplib

from nettacker.core.lib.base import BaseEngine, BaseLibrary


class Pop3Library(BaseLibrary):
    client = poplib.POP3

    def brute_force(self, host, port, username, password, timeout):
        connection = self.client(host, port=port, timeout=timeout)
        connection.user(username)
        connection.pass_(password)
        connection.quit()

        return {
            "host": host,
            "port": port,
            "username": username,
            "password": password,
        }


class Pop3Engine(BaseEngine):
    library = Pop3Library



import csv
import json
import multiprocessing
import os
import random
import string
import time
from threading import Thread
from types import SimpleNamespace

from flask import Flask, Response, abort, jsonify, make_response, render_template
from flask import request as flask_request
from werkzeug.serving import WSGIRequestHandler
from werkzeug.utils import secure_filename

from nettacker import logger
from nettacker.api.core import (
    api_key_is_valid,
    get_file,
    get_value,
    graphs,
    languages_to_country,
    mime_types,
    profiles,
    scan_methods,
)
from nettacker.api.helpers import structure
from nettacker.config import Config
from nettacker.core.app import Nettacker
from nettacker.core.die import die_failure
from nettacker.core.graph import create_compare_report
from nettacker.core.messages import messages as _
from nettacker.core.utils.common import generate_compare_filepath, now
from nettacker.database.db import (
    create_connection,
    get_logs_by_scan_id,
    get_scan_result,
    last_host_logs,
    logs_to_report_html,
    logs_to_report_json,
    search_logs,
    select_reports,
)
from nettacker.database.models import Report

# Monkey-patching the Server header to avoid exposing the actual version
WSGIRequestHandler.version_string = lambda self: "API"

log = logger.get_logger()

app = Flask(__name__, template_folder=str(Config.path.web_static_dir))
app.config.from_object(__name__)

nettacker_path_config = Config.path
nettacker_application_config = Config.settings.as_dict()
nettacker_application_config.update(Config.api.as_dict())
del nettacker_application_config["api_access_key"]


@app.errorhandler(400)
def error_400(error):
    """
    handle the 400 HTTP error

    Args:
        error: the flask error

    Returns:
        400 JSON error
    """
    return jsonify(structure(status="error", msg=error.description)), 400


@app.errorhandler(401)
def error_401(error):
    """
    handle the 401 HTTP error

    Args:
        error: the flask error

    Returns:
        401 JSON error
    """
    return jsonify(structure(status="error", msg=error.description)), 401


@app.errorhandler(403)
def error_403(error):
    """
    handle the 403 HTTP error

    Args:
        error: the flask error

    Returns:
        403 JSON error
    """
    return jsonify(structure(status="error", msg=error.description)), 403


@app.errorhandler(404)
def error_404(error):
    """
    handle the 404 HTTP error

    Args:
        error: the flask error

    Returns:
        404 JSON error
    """
    return jsonify(structure(status="error", msg=_("not_found"))), 404


@app.before_request
def limit_remote_addr():
    """
    check if IP filtering applied and API address is in whitelist

    Returns:
        None if it's in whitelist otherwise abort(403)
    """
    # IP Limitation
    if app.config["OWASP_NETTACKER_CONFIG"]["api_client_whitelisted_ips"]:
        if (
            flask_request.remote_addr
            not in app.config["OWASP_NETTACKER_CONFIG"]["api_client_whitelisted_ips"]
        ):
            abort(403, _("unauthorized_IP"))
    return


@app.after_request
def set_security_headers(response):
    """
    Add common security headers to every response.
    """
    response.headers.setdefault("Content-Security-Policy", "upgrade-insecure-requests")
    response.headers.setdefault("X-Content-Type-Options", "nosniff")
    response.headers.setdefault("X-Frame-Options", "SAMEORIGIN")
    response.headers.setdefault("X-XSS-Protection", "1; mode=block")
    response.headers.setdefault("Referrer-Policy", "no-referrer-when-downgrade")
    return response


@app.after_request
def access_log(response):
    """
    Write to the access log file if enabled.

    Args:
        response: the flask response

    Returns:
        the flask response
    """
    if app.config["OWASP_NETTACKER_CONFIG"]["api_access_log"]:
        log_request = open(app.config["OWASP_NETTACKER_CONFIG"]["api_access_log"], "ab")
        log_request.write(
            '{0} [{1}] {2} "{3} {4}" {5} {6} {7}\r\n'.format(
                flask_request.remote_addr,
                now(),
                flask_request.host,
                flask_request.method,
                flask_request.full_path,
                flask_request.user_agent,
                response.status_code,
                json.dumps(flask_request.form),
            ).encode()
        )
        log_request.close()
    return response


@app.route("/<path:path>")
def get_statics(path):
    """
    getting static files and return content mime types

    Args:
        path: path and filename

    Returns:
        file content and content type if file found otherwise abort(404)
    """
    static_types = mime_types()
    return Response(
        get_file(os.path.join(Config.path.web_static_dir, path)),
        mimetype=static_types.get(os.path.splitext(path)[1], "text/html"),
    )


@app.route("/", methods=["GET", "POST"])
def index():
    """
    index page for WebUI

    Returns:
        rendered HTML page
    """
    return render_template(
        "index.html",
        selected_modules=scan_methods(),
        profile=profiles(),
        languages=languages_to_country(),
        graphs=graphs(),
        filename=Config.settings.report_path_filename,
    )


def sanitize_report_path_filename(report_path_filename):
    """
    sanitize the report_path_filename

    Args:
        report_path_filename: the report path filename

    Returns:
        the sanitized report path filename
    """
    filename = secure_filename(os.path.basename(report_path_filename))
    if not filename:
        return False
    # Define a list or tuple of valid extensions
    VALID_EXTENSIONS = (".html", ".htm", ".txt", ".json", ".csv")
    if "." in filename:
        if filename.endswith(VALID_EXTENSIONS):
            safe_report_path = nettacker_path_config.results_dir / filename
        else:
            return False
    else:
        safe_report_path = nettacker_path_config.results_dir / filename
    if not safe_report_path.is_relative_to(nettacker_path_config.results_dir):
        return False
    return safe_report_path


@app.route("/new/scan", methods=["GET", "POST"])
def new_scan():
    """
    new scan through the API

    Returns:
        a JSON message with scan details if success otherwise a JSON error
    """
    api_key_is_valid(app, flask_request)
    form_values = dict(flask_request.form)
    # variables for future reference
    raw_report_path_filename = form_values.get("report_path_filename")
    http_header = form_values.get("http_header")
    report_path_filename = sanitize_report_path_filename(raw_report_path_filename)
    if not report_path_filename:
        return jsonify(structure(status="error", msg="Invalid report filename")), 400
    form_values["report_path_filename"] = str(report_path_filename)
    for key in nettacker_application_config:
        if key not in form_values:
            form_values[key] = nettacker_application_config[key]
    # Handle HTTP headers
    if http_header:
        form_values["http_header"] = [
            line.strip() for line in http_header.split("\n") if line.strip()
        ]
    # Handle service discovery
    form_values["skip_service_discovery"] = form_values.get("skip_service_discovery", "") == "true"
    nettacker_app = Nettacker(api_arguments=SimpleNamespace(**form_values))
    app.config["OWASP_NETTACKER_CONFIG"]["options"] = nettacker_app.arguments
    thread = Thread(target=nettacker_app.run)
    thread.start()

    return jsonify(vars(nettacker_app.arguments)), 200


@app.route("/compare/scans", methods=["POST"])
def compare_scans():
    """
    compare two scans through the API
    Returns:
        Success if the comparision is successfull and report is saved and error if not.
    """
    api_key_is_valid(app, flask_request)

    scan_id_first = get_value(flask_request, "scan_id_first")
    scan_id_second = get_value(flask_request, "scan_id_second")
    if not scan_id_first or not scan_id_second:
        return jsonify(structure(status="error", msg="Invalid Scan IDs")), 400

    compare_report_path_filename = get_value(flask_request, "compare_report_path")
    if not compare_report_path_filename:
        compare_report_path_filename = generate_compare_filepath(scan_id_first)

    compare_options = {
        "scan_compare_id": scan_id_second,
        "compare_report_path_filename": compare_report_path_filename,
    }

    try:
        result = create_compare_report(compare_options, scan_id_first)
        if result:
            return jsonify(
                structure(
                    status="success",
                    msg="scan_comparison_completed",
                )
            ), 200
        return jsonify(structure(status="error", msg="Scan ID not found")), 404
    except (FileNotFoundError, PermissionError, IOError):
        return jsonify(structure(status="error", msg="Invalid file path")), 400


@app.route("/session/check", methods=["GET"])
def session_check():
    """
    check the session if it's valid

    Returns:
        a JSON message if it's valid otherwise abort(401)
    """
    api_key_is_valid(app, flask_request)
    return jsonify(structure(status="ok", msg=_("browser_session_valid"))), 200


@app.route("/session/set", methods=["GET", "POST"])
def session_set():
    """
    set session on the browser

    Returns:
        200 HTTP response if session is valid and a set-cookie in the
        response if success otherwise abort(403)
    """
    api_key_is_valid(app, flask_request)
    res = make_response(jsonify(structure(status="ok", msg=_("browser_session_valid"))))
    res.set_cookie(
        "key",
        value=app.config["OWASP_NETTACKER_CONFIG"]["api_access_key"],
        httponly=True,
        samesite="Lax",
        secure=True,
    )
    return res


@app.route("/session/kill", methods=["GET"])
def session_kill():
    """
    unset session on the browser

    Returns:
        a 200 HTTP response with set-cookie to "expired"
        to unset the cookie on the browser
    """
    res = make_response(jsonify(structure(status="ok", msg=_("browser_session_killed"))))
    res.set_cookie("key", "", expires=0)
    return res


@app.route("/results/get_list", methods=["GET"])
def get_results():
    """
    get list of scan's results through the API

    Returns:
        an array of JSON scan's results if success otherwise abort(403)
    """
    api_key_is_valid(app, flask_request)
    page = get_value(flask_request, "page")
    if not page:
        page = 1
    return jsonify(select_reports(int(page))), 200


@app.route("/results/get", methods=["GET"])
def get_result_content():
    """
    get a result HTML/TEXT/JSON content

    Returns:
        content of the scan result
    """
    api_key_is_valid(app, flask_request)
    scan_id = get_value(flask_request, "id")
    if not scan_id:
        return jsonify(structure(status="error", msg=_("invalid_scan_id"))), 400

    try:
        filename, file_content = get_scan_result(scan_id)
    except Exception:
        return jsonify(structure(status="error", msg="database error!")), 500

    return Response(
        file_content,
        mimetype=mime_types().get(os.path.splitext(filename)[1], "text/plain"),
        headers={"Content-Disposition": "attachment;filename=" + filename.split("/")[-1]},
    )


@app.route("/results/get_json", methods=["GET"])
def get_results_json():
    """
    get host's logs through the API in JSON type

    Returns:
        an array with JSON events
    """
    api_key_is_valid(app, flask_request)
    session = create_connection()
    result_id = get_value(flask_request, "id")
    if not result_id:
        return jsonify(structure(status="error", msg=_("invalid_scan_id"))), 400
    scan_details = session.query(Report).filter(Report.id == result_id).first()
    json_object = json.dumps(get_logs_by_scan_id(scan_details.scan_unique_id))
    filename = ".".join(scan_details.report_path_filename.split(".")[:-1])[1:] + ".json"
    return Response(
        json_object,
        mimetype="application/json",
        headers={"Content-Disposition": "attachment;filename=" + filename},
    )


@app.route("/results/get_csv", methods=["GET"])
def get_results_csv():  # todo: need to fix time format
    """
    get host's logs through the API in JSON type

    Returns:
        an array with JSON events
    """
    api_key_is_valid(app, flask_request)
    session = create_connection()
    result_id = get_value(flask_request, "id")
    if not result_id:
        return jsonify(structure(status="error", msg=_("invalid_scan_id"))), 400
    scan_details = session.query(Report).filter(Report.id == result_id).first()
    data = get_logs_by_scan_id(scan_details.scan_unique_id)
    if not data:
        return jsonify(structure(status="error", msg=_("no_scan_data_found"))), 404
    keys = data[0].keys()
    filename = ".".join(scan_details.report_path_filename.split(".")[:-1])[1:] + ".csv"
    with open(filename, "w") as report_path_filename:
        dict_writer = csv.DictWriter(report_path_filename, fieldnames=keys, quoting=csv.QUOTE_ALL)
        dict_writer.writeheader()
        for event in data:
            dict_writer.writerow({key: value for key, value in event.items() if key in keys})
    with open(filename, "r") as report_path_filename:
        reader = report_path_filename.read()
    return Response(
        reader,
        mimetype="text/csv",
        headers={"Content-Disposition": "attachment;filename=" + filename},
    )


@app.route("/logs/get_list", methods=["GET"])
def get_last_host_logs():  # need to check
    """
    get list of logs through the API

    Returns:
        an array of JSON logs if success otherwise abort(403)
    """
    api_key_is_valid(app, flask_request)
    page = get_value(flask_request, "page")
    if not page:
        page = 1
    return jsonify(last_host_logs(int(page))), 200


@app.route("/logs/get_html", methods=["GET"])
def get_logs_html():  # todo: check until here - ali
    """
    get host's logs through the API in HTML type

    Returns:
        HTML report
    """
    api_key_is_valid(app, flask_request)
    target = get_value(flask_request, "target")
    return make_response(logs_to_report_html(target))


@app.route("/logs/get_json", methods=["GET"])
def get_logs():
    """
    get host's logs through the API in JSON type

    Returns:
        an array with JSON events
    """
    api_key_is_valid(app, flask_request)
    target = get_value(flask_request, "target")
    data = logs_to_report_json(target)
    json_object = json.dumps(data)
    filename = (
        "report-"
        + now(format="%Y_%m_%d_%H_%M_%S")
        + "".join(random.choice(string.ascii_lowercase) for _ in range(10))
    )
    return Response(
        json_object,
        mimetype="application/json",
        headers={"Content-Disposition": "attachment;filename=" + filename + ".json"},
    )


@app.route("/logs/get_csv", methods=["GET"])
def get_logs_csv():
    """
    get target's logs through the API in JSON type

    Returns:
        an array with JSON events
    """
    api_key_is_valid(app, flask_request)
    target = get_value(flask_request, "target")
    data = logs_to_report_json(target)
    if not data:
        return jsonify(structure(status="error", msg=_("no_scan_data_found"))), 404
    keys = data[0].keys()
    filename = (
        "report-"
        + now(format="%Y_%m_%d_%H_%M_%S")
        + "".join(random.choice(string.ascii_lowercase) for _ in range(10))
    )
    with open(filename, "w") as report_path_filename:
        dict_writer = csv.DictWriter(report_path_filename, fieldnames=keys, quoting=csv.QUOTE_ALL)
        dict_writer.writeheader()
        for event in data:
            dict_writer.writerow({key: value for key, value in event.items() if key in keys})
    with open(filename, "r") as report_path_filename:
        reader = report_path_filename.read()
    return Response(
        reader,
        mimetype="text/csv",
        headers={"Content-Disposition": f"attachment;filename={filename}.csv"},
    )


@app.route("/logs/search", methods=["GET"])
def go_for_search_logs():
    """
    search in all events

    Returns:
        an array with JSON events
    """
    api_key_is_valid(app, flask_request)
    try:
        page = int(get_value(flask_request, "page"))
        if page > 0:
            page -= 1
    except Exception:
        page = 0
    try:
        query = get_value(flask_request, "q")
    except Exception:
        query = ""
    return jsonify(search_logs(page, query)), 200


def start_api_subprocess(options):
    """
    a function to run flask in a subprocess to make kill signal in a better
    way!

    Args:
        options: all options
    """
    app.config["OWASP_NETTACKER_CONFIG"] = {
        "api_access_key": options.api_access_key,
        "api_client_whitelisted_ips": options.api_client_whitelisted_ips,
        "api_access_log": options.api_access_log,
        "api_cert": options.api_cert,
        "api_cert_key": options.api_cert_key,
        "language": options.language,
        "options": options,
    }
    try:
        if options.api_cert and options.api_cert_key:
            app.run(
                host=options.api_hostname,
                port=options.api_port,
                debug=options.api_debug_mode,
                use_reloader=False,
                ssl_context=(options.api_cert, options.api_cert_key),
                threaded=True,
            )
        else:
            app.run(
                host=options.api_hostname,
                port=options.api_port,
                debug=options.api_debug_mode,
                use_reloader=False,
                ssl_context="adhoc",
                threaded=True,
            )
    except Exception as e:
        die_failure(str(e))


def start_api_server(options):
    """
    entry point to run the API through the flask

    Args:
        options: all options
    """
    # Starting the API
    log.write_to_api_console(_("API_key").format(options.api_port, options.api_access_key))
    p = multiprocessing.Process(target=start_api_subprocess, args=(options,))
    p.start()
    # Sometimes it's take much time to terminate flask with CTRL+C
    # So It's better to use KeyboardInterrupt to terminate!
    while len(multiprocessing.active_children()) != 0:
        try:
            time.sleep(0.3)
        except KeyboardInterrupt:
            for process in multiprocessing.active_children():
                process.terminate()
            break



def start(events):
    """
    generate the d3_tree_v2_graph with events (using d3_tree_v1_graph)

    Args:
        events: events

    Returns:
        a graph in HTML
    """
    from nettacker.lib.graph.d3_tree_v1.engine import start

    return start(events).replace(
        """\t root.children.forEach(function(child){
     collapse(child);
\t });""",
        "",
    )



from sqlalchemy import create_engine

from nettacker.config import Config
from nettacker.database.models import Base


def sqlite_create_tables():
    """
    when using sqlite database, this is the function that is used to create
    the database schema for the first time when you run the nettacker module.

    """
    db_engine = create_engine(
        "sqlite:///{name}".format(**Config.db.as_dict()),
        connect_args={"check_same_thread": False},
    )
    Base.metadata.create_all(db_engine)



from sqlalchemy import create_engine, text

from nettacker.config import Config
from nettacker.database.models import Base


def mysql_create_database():
    """
    when using mysql database, this is the function that is used to create the
    database for the first time when you run the nettacker module.
    """
    engine = create_engine(
        "mysql+pymysql://{username}:{password}@{host}:{port}".format(**Config.db.as_dict())
    )

    # Earlier we were quering using the engine itself, but we first need to create a connection
    try:
        with engine.connect() as conn:
            existing_databases = conn.execute(text("SHOW DATABASES;"))
            existing_databases = [d[0] for d in existing_databases]

            if Config.db.name not in existing_databases:
                conn.execute(text("CREATE DATABASE {0} ".format(Config.db.name)))
    except Exception as e:
        print(e)


def mysql_create_tables():
    """
    when using mysql database, this is the function that is used to create the
    tables in the database for the first time when you run the nettacker module.

    Args:
        None

    Returns:
        True if success otherwise False
    """
    db_engine = create_engine(
        "mysql+pymysql://{username}:{password}@{host}:{port}/{name}".format(**Config.db.as_dict())
    )
    Base.metadata.create_all(db_engine)



import sys
from io import StringIO

import yaml

from nettacker.config import Config
from nettacker.core.utils.common import find_args_value


def application_language():
    if "-L" in sys.argv:
        language = find_args_value("-L") or "en"
    elif "--language" in sys.argv:
        language = find_args_value("--language") or "en"
    else:
        language = Config.settings.language
    if language not in get_languages():
        language = "en"
    return language


def load_yaml(filename):
    return yaml.load(StringIO(open(filename, "r").read()), Loader=yaml.FullLoader)


def get_languages():
    """
    Get available languages

    Returns:
        an array of languages
    """
    languages_list = []

    for language in Config.path.locale_dir.glob("*.yaml"):
        languages_list.append(str(language).split("/")[-1].split(".")[0])
    return list(set(languages_list))


class load_message:
    def __init__(self):
        self.language = application_language()
        self.messages = load_yaml(
            "{messages_path}/{language}.yaml".format(
                messages_path=Config.path.locale_dir, language=self.language
            )
        )
        if self.language != "en":
            self.messages_en = load_yaml(
                "{messages_path}/en.yaml".format(messages_path=Config.path.locale_dir)
            )
            for message_id in self.messages_en:
                if message_id not in self.messages:
                    self.messages[message_id] = self.messages_en[message_id]


message_cache = load_message().messages


def messages(msg_id):
    """
    load a message from message library with specified language

    Args:
        msg_id: message id

    Returns:
        the message content in the selected language if
        message found otherwise return message in English
    """
    return message_cache[str(msg_id)]



from impacket.smbconnection import SMBConnection

from nettacker.core.lib.base import BaseEngine, BaseLibrary


def create_connection(host, port):
    return SMBConnection(host, remoteHost=host, sess_port=port)


class SmbLibrary(BaseLibrary):
    def brute_force(self, *args, **kwargs):
        host = kwargs["host"]
        port = kwargs["port"]
        username = kwargs["username"]

        response = {
            "host": host,
            "port": port,
            "username": username,
        }

        domain = "."
        if "domain" in kwargs:
            domain = kwargs["domain"]
            response.update({"domain": domain})

        password = ""
        if "password" in kwargs:
            password = kwargs["password"]
            response.update({"password": password})

        lm = ""
        if "lm" in kwargs:
            lm = kwargs["lm"]
            response.update({"lm": lm})

        nt = ""
        if "nt" in kwargs:
            nt = kwargs["nt"]
            response.update({"nt": nt})

        connection = create_connection(host, port)
        connection.login(username, password, domain, lm, nt)

        return response


class SmbEngine(BaseEngine):
    library = SmbLibrary



from nettacker.config import Config

css_1 = open(Config.path.web_static_dir / "report/html_table.css").read()
json_parse_js = open(Config.path.web_static_dir / "report/json_parse.js").read()
table_end = open(Config.path.web_static_dir / "report/table_end.html").read()
table_items = open(Config.path.web_static_dir / "report/table_items.html").read()
table_title = open(Config.path.web_static_dir / "report/table_title.html").read()



import copy
import ctypes
import datetime
import hashlib
import importlib
import json
import math
import multiprocessing
import random
import re
import string
import sys
import time
from itertools import product

from nettacker import logger

log = logger.get_logger()


def replace_dependent_response(log, response_dependent):
    """The `response_dependent` is needed for `eval` below."""
    if str(log):
        key_name = re.findall(re.compile("response_dependent\\['\\S+\\]"), log)
        for i in key_name:
            try:
                key_value = eval(i)
            except Exception:
                key_value = "response dependent error"
            log = log.replace(i, " ".join(key_value))
        return log


def merge_logs_to_list(result, log_list=None):
    """Recursively extract all 'log' values from a nested dict into a flat deduplicated list.

    Args:
        result: A dict (possibly nested) containing 'log' keys to extract.
        log_list: Accumulator list for recursive calls. Defaults to a new empty list
            on each top-level call to avoid mutable default argument pitfalls.

    Returns:
        A deduplicated list of extracted log values.
    """
    if log_list is None:
        log_list = []
    if isinstance(result, dict):
        if "json_event" in list(result.keys()):
            if not isinstance(result["json_event"], dict):
                result["json_event"] = json.loads(result["json_event"])
        for i in result:
            if "log" == i:
                log_list.append(result["log"])
            else:
                merge_logs_to_list(result[i], log_list)
    return list(set(log_list))


def reverse_and_regex_condition(regex, reverse):
    if regex:
        if reverse:
            return []
        return list(set(regex))
    else:
        if reverse:
            return True
        return []


def wait_for_threads_to_finish(threads, maximum=None, terminable=False, sub_process=False):
    """Wait until all threads finish or the count drops below maximum.

    Args:
        threads: List of Thread (or Process) objects to monitor. Dead entries
            are removed in-place each iteration.
        maximum: If set, return early once fewer than *maximum* threads remain.
        terminable: If True, forcibly terminate surviving threads on KeyboardInterrupt.
        sub_process: If True, kill surviving sub-processes on KeyboardInterrupt.

    Returns:
        True when all threads completed (or fell below *maximum*),
        False if interrupted by KeyboardInterrupt.
    """
    while threads:
        try:
            threads[:] = [t for t in threads if t.is_alive()]
            if maximum and len(threads) < maximum:
                break
            time.sleep(0.01)
        except KeyboardInterrupt:
            if terminable:
                for thread in threads:
                    terminate_thread(thread)
            if sub_process:
                for thread in threads:
                    thread.kill()
            return False
    return True


def terminate_thread(thread, verbose=True):
    """
    kill a thread https://stackoverflow.com/a/15274929
    Args:
        thread: an alive thread
        verbose: verbose mode/boolean
    Returns:
        True/None
    """

    if verbose:
        log.info("killing {0}".format(thread.name))
    if not thread.is_alive():
        return

    exc = ctypes.py_object(SystemExit)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(thread.ident), exc)
    if res == 0:
        raise ValueError("nonexistent thread id")
    elif res > 1:
        # if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread.ident, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")
    return True


def get_http_header_key(http_header):
    """
    Return the HTTP header key based on the full string entered by the user
    Args:
        http_header: a string entered by the user following the -H flag
    Returns:
        1. The HTTP header key if http_header is a key-value pair
        2. The http_header itself if http_header is NOT a key_value pair (i.e. http_header is a plain string)
        3. An empty string if http_header is empty
    Example:
        http_header: "Authorization: Bearer abcdefgh"
        Returns -> "Authorization"
    """
    # Split only at the first ":"
    return http_header.split(":", 1)[0].strip()


def get_http_header_value(http_header):
    """
    Return the HTTP header value based on the full string entered by the user
    Args:
        http_header: a string entered by the user following the -H flag
    Returns:
        1. The HTTP header value if http_header is a key-value pair
        2. None if the http_header is empty or NOT a key-value pair
    Example:
        http_header: "Authorization: Bearer abcdefgh"
        Returns -> "Bearer abcdefgh"
    """
    if not http_header or ":" not in http_header:
        return None
    # Split only at the first ":"
    value = http_header.split(":", 1)[1].strip()
    return value if value else None


def remove_sensitive_header_keys(event):
    """
    Removes the sensitive headers that the user might add
    Args:
        event: The json event which contains the headers
    Returns:
        event: The json event without the sensitive headers
    """
    from nettacker.config import sensitive_headers

    if not isinstance(event, dict):
        return event

    if "headers" in event:
        if not isinstance(event["headers"], dict):
            return event
        for key in list(event["headers"].keys()):
            if key.lower() in sensitive_headers:
                del event["headers"][key]

    return event


def find_args_value(args_name):
    try:
        return sys.argv[sys.argv.index(args_name) + 1]
    except Exception:
        return None


def set_nested_value(d, key_path, value):
    keys = [k for k in key_path.split("/") if k]
    for key in keys[:-1]:
        d = d[key]
    d[keys[-1]] = value


def generate_new_sub_steps(sub_steps, data_matrix, arrays):
    original_sub_steps = copy.deepcopy(sub_steps)
    steps_array = []
    array_names = list(arrays.keys())
    for array in data_matrix:
        for i, array_name in enumerate(array_names):
            set_nested_value(original_sub_steps, array_name, array[i])
        steps_array.append(copy.deepcopy(original_sub_steps))
    return steps_array


def find_repeaters(sub_content, root, arrays):
    if isinstance(sub_content, dict) and "nettacker_fuzzer" not in sub_content:
        temporary_content = copy.deepcopy(sub_content)
        original_root = root
        for key in sub_content:
            root = original_root
            root += key + "/"
            temporary_content[key], _root, arrays = find_repeaters(sub_content[key], root, arrays)
        sub_content = copy.deepcopy(temporary_content)
        root = original_root
    if (not isinstance(sub_content, (bool, int, float))) and (
        isinstance(sub_content, list) or "nettacker_fuzzer" in sub_content
    ):
        arrays[root] = sub_content

    return (sub_content, root, arrays) if root != "" else arrays


class value_to_class:
    def __init__(self, value):
        self.value = value


def class_to_value(arrays):
    original_arrays = copy.deepcopy(arrays)
    array_index = 0
    for array in arrays:
        value_index = 0
        for value in array:
            if isinstance(value, value_to_class):
                original_arrays[array_index][value_index] = value.value
            value_index += 1
        array_index += 1
    return original_arrays


def generate_and_replace_md5(content):
    # todo: make it betetr and document it
    md5_content = content.split("NETTACKER_MD5_GENERATOR_START")[1].split(
        "NETTACKER_MD5_GENERATOR_STOP"
    )[0]
    md5_content_backup = md5_content
    if isinstance(md5_content, str):
        md5_content = md5_content.encode()
    md5_hash = hashlib.md5(md5_content).hexdigest()
    return content.replace(
        "NETTACKER_MD5_GENERATOR_START" + md5_content_backup + "NETTACKER_MD5_GENERATOR_STOP",
        md5_hash,
    )


def generate_target_groups(targets, set_hardware_usage):
    """
    Split a list of targets into smaller sublists based on a specified size.
    """
    if not targets:
        return targets

    targets_total = len(targets)
    split_size = min(set_hardware_usage, targets_total)

    # Calculate the size of each chunk.
    chunk_size = (targets_total + split_size - 1) // split_size

    return [targets[i : i + chunk_size] for i in range(0, targets_total, chunk_size)]


def arrays_to_matrix(arrays):
    """
    Generate a Cartesian product of input arrays as a list of lists.
    """
    return [list(item) for item in product(*[arrays[array_name] for array_name in arrays])]


def string_to_bytes(string):
    return string.encode()


AVAILABLE_DATA_FUNCTIONS = {
    "passwords": {"read_from_file"},
    "paths": {"read_from_file"},
    "urls": {"read_from_file"},
}


def fuzzer_function_read_file_as_array(filename):
    from nettacker.config import PathConfig

    return open(PathConfig().payloads_dir / filename).read().split("\n")


def apply_data_functions(data):
    original_data = copy.deepcopy(data)
    for item in data:
        if item not in AVAILABLE_DATA_FUNCTIONS:
            continue

        for fn_name in data[item]:
            if fn_name in AVAILABLE_DATA_FUNCTIONS[item]:
                fn = getattr(importlib.import_module("nettacker.core.fuzzer"), fn_name)
                if fn is not None:
                    original_data[item] = fn(data[item][fn_name])

    return original_data


ALLOWED_INTERCEPTORS = {
    "generate_and_replace_md5": generate_and_replace_md5,
}


def fuzzer_repeater_perform(arrays):
    original_arrays = copy.deepcopy(arrays)
    for array_name in arrays:
        if "nettacker_fuzzer" not in arrays[array_name]:
            continue

        data = arrays[array_name]["nettacker_fuzzer"]["data"]
        data_matrix = arrays_to_matrix(apply_data_functions(data))
        prefix = arrays[array_name]["nettacker_fuzzer"]["prefix"]
        input_format = arrays[array_name]["nettacker_fuzzer"]["input_format"]
        interceptors = copy.deepcopy(arrays[array_name]["nettacker_fuzzer"]["interceptors"])
        if interceptors:
            interceptors = interceptors.split(",")
        suffix = arrays[array_name]["nettacker_fuzzer"]["suffix"]
        processed_array = []

        for sub_data in data_matrix:
            formatted_data = {}
            index_input = 0
            for value in sub_data:
                formatted_data[list(data.keys())[index_input]] = value
                index_input += 1

            interceptors_function_processed = input_format.format(**formatted_data)

            if interceptors:
                for interceptor in interceptors:
                    if interceptor not in ALLOWED_INTERCEPTORS:
                        raise ValueError(f"Interceptor '{interceptor}' is not allowed")
                    interceptors_function_processed = ALLOWED_INTERCEPTORS[interceptor](
                        interceptors_function_processed
                    )

            processed_sub_data = interceptors_function_processed
            if prefix:
                processed_sub_data = prefix + processed_sub_data
            if suffix:
                processed_sub_data = processed_sub_data + suffix
            processed_array.append(copy.deepcopy(processed_sub_data))
        original_arrays[array_name] = processed_array

    return original_arrays


def expand_module_steps(content):
    return [expand_protocol(x) for x in copy.deepcopy(content)]


def expand_protocol(protocol):
    protocol["steps"] = [expand_step(x) for x in protocol["steps"]]
    return protocol


def expand_step(step):
    arrays = fuzzer_repeater_perform(find_repeaters(step, "", {}))
    if arrays:
        return generate_new_sub_steps(step, class_to_value(arrays_to_matrix(arrays)), arrays)
    else:
        # Minimum 1 step in array
        return [step]


def generate_random_token(length=10):
    return "".join(random.choice(string.ascii_lowercase) for _ in range(length))


def now(format="%Y-%m-%d %H:%M:%S"):
    """
    get now date and time
    Args:
        format: the date and time model, default is "%Y-%m-%d %H:%M:%S"

    Returns:
        the date and time of now
    """
    return datetime.datetime.now().strftime(format)


def select_maximum_cpu_core(mode):
    cpu_count = multiprocessing.cpu_count()

    if cpu_count - 1 == 0:
        return 1

    mode_core_mapping = {
        "maximum": cpu_count - 1,
        "high": cpu_count / 2,
        "normal": cpu_count / 4,
        "low": cpu_count / 8,
    }
    rounded = math.ceil if mode == "high" else math.floor

    return int(max((rounded(mode_core_mapping.get(mode, 1)), 1)))


def sort_dictionary(dictionary):
    etc_flag = "..." in dictionary
    if etc_flag:
        del dictionary["..."]
    sorted_dictionary = {}
    for key in sorted(dictionary):
        sorted_dictionary[key] = dictionary[key]
    if etc_flag:
        sorted_dictionary["..."] = {}
    return sorted_dictionary


def sanitize_path(path):
    """
    Sanitize the file path to preven unathorized access
    Args:
        path: filepath(user input)

    Returns:
        sanitized_path
    """
    return "_".join(
        [
            component
            for component in re.split(r"[/\\]", path)
            if re.match(r"^[a-zA-Z0-9_-]+(\.[a-zA-Z0-9_-]+)?$", component)
        ]
    )


def generate_compare_filepath(scan_id):
    return "/report_compare_{date_time}_{scan_id}.json".format(
        date_time=now(format="%Y_%m_%d_%H_%M_%S"),
        scan_id=scan_id,
    )



import logging
import socket
import ssl
from datetime import datetime, timezone

from OpenSSL import crypto

from nettacker.core.lib.base import BaseEngine, BaseLibrary

log = logging.getLogger(__name__)


def is_weak_hash_algo(algo):
    algo = algo.lower()
    for unsafe_algo in ("md2", "md4", "md5", "sha1"):
        if unsafe_algo in algo:
            return True
    return False


def create_socket_connection(context, host, port, timeout):
    socket_connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_connection.settimeout(timeout)
    socket_connection.connect((host, port))
    socket_connection = context.wrap_socket(socket_connection, server_hostname=host)
    return socket_connection


def is_weak_ssl_version(host, port, timeout):
    def test_ssl_version(host, port, timeout, ssl_version=None):
        try:
            context = ssl.SSLContext(ssl_version)
            socket_connection = create_socket_connection(context, host, port, timeout)
            return socket_connection.version()

        except ssl.SSLError:
            return False

        except (socket.timeout, ConnectionRefusedError, ConnectionResetError):
            return None

    ssl_versions = (
        ssl.PROTOCOL_TLS_CLIENT,  # TLS 1.3
        ssl.PROTOCOL_TLSv1_2,
        ssl.PROTOCOL_TLSv1_1,
        ssl.PROTOCOL_TLSv1,
    )
    supported_versions = []
    lowest_version = ""
    for ssl_version in ssl_versions:
        version = test_ssl_version(host, port, timeout, ssl_version=ssl_version)
        if version:
            lowest_version = version
            supported_versions.append(version)

    return supported_versions, lowest_version not in {"TLSv1.2", "TLSv1.3"}


def is_weak_cipher_suite(host, port, timeout):
    def test_single_cipher(host, port, cipher, timeout):
        try:
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            context.set_ciphers(cipher)
            create_socket_connection(context, host, port, timeout)
            return True

        except ssl.SSLError:
            return False

        except (socket.timeout, ConnectionRefusedError, ConnectionResetError):
            return None

    cipher_suites = [
        "HIGH",  # OpenSSL cipher strings
        "MEDIUM",
        "LOW",
        "EXP",
        "eNULL",
        "aNULL",
        "RC4",
        "DES",
        "MD5",
        "SHA1",
        "DH",
        "ADH",
        "DHE",
        "ECDH",
        "ECDHE",
        "TLSv1",
        "TLSv1.1",
        "TLSv1.2",
        "TLSv1.3",
    ]

    supported_ciphers = []
    for cipher in cipher_suites:
        if test_single_cipher(host, port, cipher, timeout):
            supported_ciphers.append(cipher)

    weak_ciphers = {"LOW", "EXP", "eNULL", "aNULL", "RC4", "DES", "MD5", "DH", "ADH"}
    for cipher in supported_ciphers:
        if cipher in weak_ciphers:
            return supported_ciphers, True

    return supported_ciphers, False


def create_tcp_socket(host, port, timeout):
    try:
        socket_connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_connection.settimeout(timeout)
        socket_connection.connect((host, port))
        ssl_flag = False
    except ConnectionRefusedError:
        return None

    try:
        socket_connection = ssl.wrap_socket(socket_connection)
        ssl_flag = True
    except Exception:
        socket_connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_connection.settimeout(timeout)
        socket_connection.connect((host, port))

    return socket_connection, ssl_flag


def get_cert_info(cert):
    x509 = crypto.load_certificate(crypto.FILETYPE_PEM, cert)
    weak_signing_algo = is_weak_hash_algo(str(x509.get_signature_algorithm()))
    cert_expires = datetime.strptime(x509.get_notAfter().decode("utf-8"), "%Y%m%d%H%M%S%z")
    cert_activation = datetime.strptime(x509.get_notBefore().decode("utf-8"), "%Y%m%d%H%M%S%z")
    issuer_str = ", ".join(
        f"{name.decode()}={value.decode()}" for name, value in x509.get_issuer().get_components()
    )
    subject_str = ", ".join(
        f"{name.decode()}={value.decode()}" for name, value in x509.get_subject().get_components()
    )
    return {
        "expired": x509.has_expired(),
        "self_signed": issuer_str == subject_str,
        "issuer": issuer_str,
        "subject": subject_str,
        "signing_algo": str(x509.get_signature_algorithm()),
        "weak_signing_algo": weak_signing_algo,
        "activation_date": cert_activation.strftime("%Y-%m-%d"),
        "not_activated": (cert_activation - datetime.now(timezone.utc)).days > 0,
        "expiration_date": cert_expires.strftime("%Y-%m-%d"),
        "expiring_soon": (cert_expires - datetime.now(timezone.utc)).days < 30,
    }


class SslLibrary(BaseLibrary):
    def ssl_certificate_scan(self, host, port, timeout):
        tcp_socket = create_tcp_socket(host, port, timeout)
        if tcp_socket is None:
            return None

        socket_connection, ssl_flag = tcp_socket
        peer_name = socket_connection.getpeername()
        scan_info = {
            "ssl_flag": ssl_flag,
            "peer_name": peer_name,
            "service": socket.getservbyport(int(port)),
        }

        if ssl_flag:
            cert = ssl.get_server_certificate((host, port))
            cert_info = get_cert_info(cert)
            scan_info = cert_info | scan_info
            return scan_info

        return scan_info

    def ssl_version_and_cipher_scan(self, host, port, timeout):
        tcp_socket = create_tcp_socket(host, port, timeout)
        if tcp_socket is None:
            return None

        socket_connection, ssl_flag = tcp_socket
        peer_name = socket_connection.getpeername()

        if ssl_flag:
            try:
                cert = ssl.get_server_certificate((host, port))
            except ssl.SSLError:
                cert = None
            except socket.gaierror:
                cert = None
            cert_info = get_cert_info(cert) if cert else None
            ssl_ver, weak_version = is_weak_ssl_version(host, port, timeout)
            cipher_suite, weak_cipher_suite = is_weak_cipher_suite(host, port, timeout)

            return {
                "ssl_version": ssl_ver,
                "weak_version": weak_version,
                "cipher_suite": cipher_suite,
                "weak_cipher_suite": weak_cipher_suite,
                "issuer": cert_info["issuer"] if cert_info else "NA",
                "subject": cert_info["subject"] if cert_info else "NA",
                "expiration_date": cert_info["expiration_date"] if cert_info else "NA",
                "ssl_flag": ssl_flag,
                "peer_name": peer_name,
                "service": socket.getservbyport(int(port)),
            }

        return {
            "ssl_flag": ssl_flag,
            "service": socket.getservbyport(int(port)),
            "peer_name": peer_name,
        }


class SslEngine(BaseEngine):
    library = SslLibrary

    def response_conditions_matched(self, sub_step, response):
        conditions = sub_step["response"]["conditions"]
        condition_type = sub_step["response"]["condition_type"]
        condition_results = {}
        if sub_step["method"] in {
            "ssl_certificate_scan",
            "ssl_version_and_cipher_scan",
        }:
            if response and response["ssl_flag"]:
                for condition in conditions:
                    if "grouped_conditions" in condition:
                        gc_type = conditions[condition]["condition_type"]
                        gc_conditions = conditions[condition]["conditions"]
                        gc_condition_results = {}
                        for gc_condition in gc_conditions:
                            if (
                                gc_conditions[gc_condition]["reverse"]
                                and not response[gc_condition]
                            ):
                                gc_condition_results[gc_condition] = not response[gc_condition]

                            elif (
                                not gc_conditions[gc_condition]["reverse"]
                                and response[gc_condition]
                            ):
                                gc_condition_results[gc_condition] = response[gc_condition]

                        if gc_type == "and":
                            gc_condition_results = (
                                gc_condition_results
                                if len(gc_condition_results) == len(gc_conditions)
                                else {}
                            )

                        condition_results.update(gc_condition_results)

                    elif (conditions[condition]["reverse"] and not response[condition]) or (
                        not conditions[condition]["reverse"] and response[condition]
                    ):
                        condition_results[condition] = True

                if condition_type == "and":
                    return condition_results if len(condition_results) == len(conditions) else []
                if condition_type == "or":
                    return condition_results if condition_results else []
                return []

        return []

    def apply_extra_data(self, sub_step, response):
        sub_step["response"]["ssl_flag"] = (
            response["ssl_flag"] if isinstance(response, dict) else False
        )
        sub_step["response"]["conditions_results"] = self.response_conditions_matched(
            sub_step, response
        )



import copy
import json
import os
import shutil
import socket
import sys
from threading import Thread

import multiprocess

from nettacker import logger
from nettacker.config import Config, version_info
from nettacker.core.arg_parser import ArgParser
from nettacker.core.die import die_failure
from nettacker.core.graph import create_compare_report, create_report
from nettacker.core.ip import (
    generate_ip_range,
    get_ip_range,
    is_ipv4_cidr,
    is_ipv4_range,
    is_ipv6_cidr,
    is_ipv6_range,
    is_single_ipv4,
    is_single_ipv6,
)
from nettacker.core.messages import messages as _
from nettacker.core.module import Module
from nettacker.core.socks_proxy import set_socks_proxy
from nettacker.core.utils import common as common_utils
from nettacker.core.utils.common import wait_for_threads_to_finish
from nettacker.database.db import find_events, remove_old_logs
from nettacker.database.mysql import mysql_create_database, mysql_create_tables
from nettacker.database.postgresql import postgres_create_database
from nettacker.database.sqlite import sqlite_create_tables
from nettacker.logger import TerminalCodes

log = logger.get_logger()


class Nettacker(ArgParser):
    def __init__(self, api_arguments=None):
        if not api_arguments:
            self.print_logo()
        self.check_dependencies()

        log.info(_("scan_started"))
        super().__init__(api_arguments=api_arguments)

    @staticmethod
    def print_logo():
        """
        OWASP Nettacker Logo
        """
        log.write_to_api_console(
            open(Config.path.logo_file)
            .read()
            .format(
                cyan=TerminalCodes.CYAN.value,
                red=TerminalCodes.RED.value,
                rst=TerminalCodes.RESET.value,
                v1=version_info()[0],
                v2=version_info()[1],
                yellow=TerminalCodes.YELLOW.value,
            )
        )
        log.reset_color()

    def check_dependencies(self):
        if sys.platform not in {"darwin", "freebsd13", "freebsd14", "freebsd15", "linux"}:
            die_failure(_("error_platform"))

        try:
            Config.path.tmp_dir.mkdir(exist_ok=True, parents=True)
            Config.path.results_dir.mkdir(exist_ok=True, parents=True)
        except PermissionError:
            die_failure("Cannot access the directory {0}".format(Config.path.tmp_dir))

        if Config.db.engine == "sqlite":
            try:
                if not Config.path.new_database_file.exists():
                    Config.path.new_database_file.parent.mkdir(parents=True, exist_ok=True)
                    if Config.path.old_database_file.exists():
                        shutil.copy(Config.path.old_database_file, Config.path.new_database_file)
                        log.warn("Database files migrated from .data to .nettacker ...")
                    else:
                        sqlite_create_tables()
            except PermissionError:
                die_failure("cannot access the directory {0}".format(Config.path.home_dir))
        elif Config.db.engine == "mysql":
            try:
                mysql_create_database()
                mysql_create_tables()
            except Exception:
                die_failure(_("database_connection_failed"))
        elif Config.db.engine == "postgres":
            try:
                postgres_create_database()
            except Exception:
                die_failure(_("database_connection_failed"))
        else:
            die_failure(_("invalid_database"))

    def expand_targets(self, scan_id):
        """
        determine targets.

        Args:
            options: all options
            scan_id: unique scan identifier

        Returns:
            a generator
        """
        targets = []
        base_path = ""
        for target in self.arguments.targets:
            if "://" in target:
                try:
                    if not target.split("://")[1].split("/")[1]:
                        base_path = ""
                    else:
                        base_path = "/".join(target.split("://")[1].split("/")[1:])
                        if base_path[-1] != "/":
                            base_path += "/"
                except IndexError:
                    base_path = ""
                # remove url proto; uri; port
                target = target.split("://")[1].split("/")[0].split(":")[0]
                targets.append(target)
            # single IPs
            elif is_single_ipv4(target) or is_single_ipv6(target):
                if self.arguments.scan_ip_range:
                    targets += get_ip_range(target)
                else:
                    targets.append(target)
            # IP ranges
            elif (
                is_ipv4_range(target)
                or is_ipv6_range(target)
                or is_ipv4_cidr(target)
                or is_ipv6_cidr(target)
            ):
                targets += generate_ip_range(target)
            # domains probably
            else:
                targets.append(target)
        self.arguments.targets = targets
        self.arguments.url_base_path = base_path

        # subdomain_scan
        if self.arguments.scan_subdomains:
            selected_modules = self.arguments.selected_modules
            self.arguments.selected_modules = ["subdomain_scan"]
            self.start_scan(scan_id)
            self.arguments.selected_modules = selected_modules
            if "subdomain_scan" in self.arguments.selected_modules:
                self.arguments.selected_modules.remove("subdomain_scan")

            for target in copy.deepcopy(self.arguments.targets):
                for row in find_events(target, "subdomain_scan", scan_id):
                    for sub_domain in json.loads(row)["response"]["conditions_results"]["content"]:
                        if sub_domain not in self.arguments.targets:
                            self.arguments.targets.append(sub_domain)
        # icmp_scan
        if self.arguments.ping_before_scan:
            if os.geteuid() == 0:
                selected_modules = self.arguments.selected_modules
                self.arguments.selected_modules = ["icmp_scan"]
                self.start_scan(scan_id)
                self.arguments.selected_modules = selected_modules
                if "icmp_scan" in self.arguments.selected_modules:
                    self.arguments.selected_modules.remove("icmp_scan")
                self.arguments.targets = self.filter_target_by_event(targets, scan_id, "icmp_scan")
            else:
                log.warn(_("icmp_need_root_access"))
                if "icmp_scan" in self.arguments.selected_modules:
                    self.arguments.selected_modules.remove("icmp_scan")
        # port_scan
        if not self.arguments.skip_service_discovery:
            self.arguments.skip_service_discovery = True
            selected_modules = self.arguments.selected_modules
            self.arguments.selected_modules = ["port_scan"]
            self.start_scan(scan_id)
            self.arguments.selected_modules = selected_modules
            if "port_scan" in self.arguments.selected_modules:
                self.arguments.selected_modules.remove("port_scan")
            self.arguments.targets = self.filter_target_by_event(targets, scan_id, "port_scan")
            self.arguments.skip_service_discovery = False
        return list(set(self.arguments.targets))

    def filter_target_by_event(self, targets, scan_id, module_name):
        for target in copy.deepcopy(targets):
            if not find_events(target, module_name, scan_id):
                targets.remove(target)
        return targets

    def run(self):
        """
        preparing for attacks and managing multi-processing for host

        Args:
            options: all options

        Returns:
            True when it ends
        """
        scan_id = common_utils.generate_random_token(32)
        log.info("ScanID: {0}".format(scan_id))
        log.info(_("regrouping_targets"))
        # find total number of targets + types + expand (subdomain, IPRanges, etc)
        # optimize CPU usage
        self.arguments.targets = self.expand_targets(scan_id)
        if not self.arguments.targets:
            log.info(_("no_live_service_found"))
            return True
        exit_code = self.start_scan(scan_id)
        create_report(self.arguments, scan_id)
        if self.arguments.scan_compare_id is not None:
            create_compare_report(self.arguments, scan_id)
        log.info("ScanID: {0} ".format(scan_id) + _("done"))

        return exit_code

    def start_scan(self, scan_id):
        target_groups = common_utils.generate_target_groups(
            self.arguments.targets, self.arguments.set_hardware_usage
        )
        log.info(_("removing_old_db_records"))

        for target_group in target_groups:
            for target in target_group:
                for module_name in self.arguments.selected_modules:
                    remove_old_logs(
                        {
                            "target": target,
                            "module_name": module_name,
                            "scan_id": scan_id,
                            "scan_compare_id": self.arguments.scan_compare_id,
                        }
                    )

        for _i in range(target_groups.count([])):
            target_groups.remove([])

        log.info(_("start_multi_process").format(len(self.arguments.targets), len(target_groups)))
        active_processes = []
        for t_id, target_groups in enumerate(target_groups):
            process = multiprocess.Process(
                target=self.scan_target_group, args=(target_groups, scan_id, t_id)
            )
            process.start()
            active_processes.append(process)

        return wait_for_threads_to_finish(active_processes, sub_process=True)

    def scan_target(
        self,
        target,
        module_name,
        scan_id,
        process_number,
        thread_number,
        total_number_threads,
    ):
        options = copy.deepcopy(self.arguments)

        socket.socket, socket.getaddrinfo = set_socks_proxy(options.socks_proxy)
        module = Module(
            module_name,
            options,
            target,
            scan_id,
            process_number,
            thread_number,
            total_number_threads,
        )
        module.load()
        module.generate_loops()
        module.sort_loops()
        module.start()

        log.verbose_event_info(
            _("finished_parallel_module_scan").format(
                process_number, module_name, target, thread_number, total_number_threads
            )
        )

        return os.EX_OK

    def scan_target_group(self, targets, scan_id, process_number):
        active_threads = []
        log.verbose_event_info(_("single_process_started").format(process_number))
        total_number_of_modules = len(targets) * len(self.arguments.selected_modules)
        total_number_of_modules_counter = 1

        for target in targets:
            for module_name in self.arguments.selected_modules:
                thread = Thread(
                    target=self.scan_target,
                    args=(
                        target,
                        module_name,
                        scan_id,
                        process_number,
                        total_number_of_modules_counter,
                        total_number_of_modules,
                    ),
                )
                thread.name = f"{target} -> {module_name}"
                thread.start()
                log.verbose_event_info(
                    _("start_parallel_module_scan").format(
                        process_number,
                        module_name,
                        target,
                        total_number_of_modules_counter,
                        total_number_of_modules,
                    )
                )
                total_number_of_modules_counter += 1
                active_threads.append(thread)
                if not wait_for_threads_to_finish(
                    active_threads, self.arguments.parallel_module_scan, True
                ):
                    return False
        wait_for_threads_to_finish(active_threads, maximum=None, terminable=True)
        return True



#!/usr/bin/env python

import copy
import logging
import os
import re
import select
import socket
import ssl
import struct
import time

from nettacker.core.lib.base import BaseEngine, BaseLibrary
from nettacker.core.utils.common import replace_dependent_response, reverse_and_regex_condition

log = logging.getLogger(__name__)


def create_tcp_socket(host, port, timeout):
    try:
        socket_connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_connection.settimeout(timeout)
        socket_connection.connect((host, port))
        ssl_flag = False
    except ConnectionRefusedError:
        return None

    try:
        socket_connection = ssl.wrap_socket(socket_connection)
        ssl_flag = True
    except Exception:
        socket_connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_connection.settimeout(timeout)
        socket_connection.connect((host, port))
    # finally:
    #     socket_connection.shutdown()

    return socket_connection, ssl_flag


class SocketLibrary(BaseLibrary):
    def tcp_connect_only(self, host, port, timeout):
        tcp_socket = create_tcp_socket(host, port, timeout)
        if tcp_socket is None:
            return None

        socket_connection, ssl_flag = tcp_socket
        peer_name = socket_connection.getpeername()
        socket_connection.close()

        try:
            service = socket.getservbyport(port)
        except OSError:
            service = "unknown"

        return {
            "peer_name": peer_name,
            "service": service,
            "ssl_flag": ssl_flag,
        }

    def tcp_connect_send_and_receive(self, host, port, timeout):
        tcp_socket = create_tcp_socket(host, port, timeout)
        if tcp_socket is None:
            return None

        socket_connection, ssl_flag = tcp_socket
        peer_name = socket_connection.getpeername()
        try:
            socket_connection.send(b"ABC\x00\r\n\r\n\r\n" * 10)
            response = socket_connection.recv(1024 * 1024 * 10)
            socket_connection.close()
        # except ConnectionRefusedError:
        #     return None
        except Exception:
            try:
                socket_connection.close()
                response = b""
            except Exception:
                response = b""

        try:
            service = socket.getservbyport(port)
        except OSError:
            service = "unknown"

        return {
            "peer_name": peer_name,
            "response": response.decode(errors="ignore"),
            "service": service,
            "ssl_flag": ssl_flag,
        }

    def socket_icmp(self, host, timeout):
        """
        A pure python ping implementation using raw socket.
        Note that ICMP messages can only be sent from processes running as root.
        Derived from ping.c distributed in Linux's netkit. That code is
        copyright (c) 1989 by The Regents of the University of California.
        That code is in turn derived from code written by Mike Muuss of the
        US Army Ballistic Research Laboratory in December, 1983 and
        placed in the public domain. They have my thanks.
        Bugs are naturally mine. I'd be glad to hear about them. There are
        certainly word - size dependenceies here.
        Copyright (c) Matthew Dixon Cowles, <http://www.visi.com/~mdc/>.
        Distributable under the terms of the GNU General Public License
        version 2. Provided with no warranties of any sort.
        Original Version from Matthew Dixon Cowles:
          -> ftp://ftp.visi.com/users/mdc/ping.py
        Rewrite by Jens Diemer:
          -> http://www.python-forum.de/post-69122.html#69122
        Rewrite by George Notaras:
          -> http://www.g-loaded.eu/2009/10/30/python-ping/
        Fork by Pierre Bourdon:
          -> http://bitbucket.org/delroth/python-ping/
        Revision history
        ~~~~~~~~~~~~~~~~
        November 22, 1997
        -----------------
        Initial hack. Doesn't do much, but rather than try to guess
        what features I (or others) will want in the future, I've only
        put in what I need now.
        December 16, 1997
        -----------------
        For some reason, the checksum bytes are in the wrong order when
        this is run under Solaris 2.X for SPARC but it works right under
        Linux x86. Since I don't know just what's wrong, I'll swap the
        bytes always and then do an htons().
        December 4, 2000
        ----------------
        Changed the struct.pack() calls to pack the checksum and ID as
        unsigned. My thanks to Jerome Poincheval for the fix.
        May 30, 2007
        ------------
        little rewrite by Jens Diemer:
         -  change socket asterisk import to a normal import
         -  replace time.time() with time.clock()
         -  delete "return None" (or change to "return" only)
         -  in checksum() rename "str" to "source_string"
        November 8, 2009
        ----------------
        Improved compatibility with GNU/Linux systems.
        Fixes by:
         * George Notaras -- http://www.g-loaded.eu
        Reported by:
         * Chris Hallman -- http://cdhallman.blogspot.com
        Changes in this release:
         - Re-use time.time() instead of time.clock(). The 2007 implementation
           worked only under Microsoft Windows. Failed on GNU/Linux.
           time.clock() behaves differently under the two OSes[1].
        [1] http://docs.python.org/library/time.html#time.clock
        September 25, 2010
        ------------------
        Little modifications by Georgi Kolev:
         -  Added quiet_ping function.
         -  returns percent lost packages, max round trip time, avrg round trip
            time
         -  Added packet size to verbose_ping & quiet_ping functions.
         -  Bump up version to 0.2
        ------------------
        5 Aug 2021 - Modified by Ali Razmjoo Qalaei (Reformat the code and more human readable)
        """
        icmp_socket = socket.getprotobyname("icmp")
        socket_connection = socket.socket(socket.AF_INET, socket.SOCK_RAW, icmp_socket)
        random_integer = os.getpid() & 0xFFFF
        icmp_echo_request = 8
        # Make a dummy header with a 0 checksum.
        dummy_checksum = 0
        header = struct.pack("bbHHh", icmp_echo_request, 0, dummy_checksum, random_integer, 1)
        data = (
            struct.pack("d", time.time())
            + struct.pack("d", time.time())
            + str((76 - struct.calcsize("d")) * "Q").encode()
        )  # packet size = 76 (removed 8 bytes size of header)
        source_string = header + data
        # Calculate the checksum on the data and the dummy header.
        calculate_data = 0
        max_size = (len(source_string) / 2) * 2
        counter = 0
        while counter < max_size:
            calculate_data += source_string[counter + 1] * 256 + source_string[counter]
            calculate_data = calculate_data & 0xFFFFFFFF  # Necessary?
            counter += 2

        if max_size < len(source_string):
            calculate_data += source_string[len(source_string) - 1]
            calculate_data = calculate_data & 0xFFFFFFFF  # Necessary?

        calculate_data = (calculate_data >> 16) + (calculate_data & 0xFFFF)
        calculate_data = calculate_data + (calculate_data >> 16)
        calculated_data = ~calculate_data & 0xFFFF

        # Swap bytes. Bugger me if I know why.
        dummy_checksum = calculated_data >> 8 | (calculated_data << 8 & 0xFF00)

        header = struct.pack(
            "bbHHh",
            icmp_echo_request,
            0,
            socket.htons(dummy_checksum),
            random_integer,
            1,
        )
        socket_connection.sendto(
            header + data, (socket.gethostbyname(host), 1)
        )  # Don't know about the 1

        while True:
            started_select = time.time()
            what_ready = select.select([socket_connection], [], [], timeout)
            how_long_in_select = time.time() - started_select
            if not what_ready[0]:  # Timeout
                break
            time_received = time.time()
            received_packet, address = socket_connection.recvfrom(1024)
            icmp_header = received_packet[20:28]
            (
                packet_type,
                packet_code,
                packet_checksum,
                packet_id,
                packet_sequence,
            ) = struct.unpack("bbHHh", icmp_header)
            if packet_id == random_integer:
                packet_bytes = struct.calcsize("d")
                time_sent = struct.unpack("d", received_packet[28 : 28 + packet_bytes])[0]
                delay = time_received - time_sent
                break

            timeout = timeout - how_long_in_select
            if timeout <= 0:
                break
        socket_connection.close()
        return {"host": host, "response_time": delay, "ssl_flag": False}


class SocketEngine(BaseEngine):
    library = SocketLibrary

    def response_conditions_matched(self, sub_step, response):
        conditions = sub_step["response"]["conditions"].get(
            "service", sub_step["response"]["conditions"]
        )
        condition_type = sub_step["response"]["condition_type"]
        condition_results = {}
        if sub_step["method"] == "tcp_connect_only":
            return response
        if sub_step["method"] == "tcp_connect_send_and_receive":
            if response:
                for condition in conditions:
                    regex = re.findall(
                        re.compile(conditions[condition]["regex"]),
                        response["response"]
                        if condition != "open_port"
                        else str(response["peer_name"][1]),
                    )
                    reverse = conditions[condition]["reverse"]
                    condition_results[condition] = reverse_and_regex_condition(regex, reverse)

                    if condition_results[condition]:
                        default_service = response["service"]
                        ssl_flag = response["ssl_flag"]
                        matched_regex = condition_results[condition]

                        log_response = {
                            "running_service": condition,
                            "matched_regex": matched_regex,
                            "default_service": default_service,
                            "ssl_flag": ssl_flag,
                        }
                        condition_results["service"] = [str(log_response)]
                for condition in copy.deepcopy(condition_results):
                    if not condition_results[condition]:
                        del condition_results[condition]

                if "open_port" in condition_results and len(condition_results) > 1:
                    del condition_results["open_port"]
                    del conditions["open_port"]
                if condition_type.lower() == "and":
                    return condition_results if len(condition_results) == len(conditions) else []
                if condition_type.lower() == "or":
                    if sub_step["response"].get("log", False):
                        condition_results["log"] = sub_step["response"]["log"]
                        if "response_dependent" in condition_results["log"]:
                            condition_results["log"] = replace_dependent_response(
                                condition_results["log"], condition_results
                            )
                    return condition_results if condition_results else []
                return []
        if sub_step["method"] == "socket_icmp":
            return response
        return []

    def apply_extra_data(self, sub_step, response):
        sub_step["response"]["ssl_flag"] = (
            response["ssl_flag"] if isinstance(response, dict) else False
        )
        sub_step["response"]["conditions_results"] = self.response_conditions_matched(
            sub_step, response
        )



import sys

from nettacker import logger

log = logger.get_logger()


def die_success():
    """
    exit the framework with code 0
    """
    log.reset_color()
    sys.exit(0)


def die_failure(msg):
    """
    exit the framework with code 1

    Args:
        msg: the error message
    """

    log.error(msg)
    log.reset_color()
    sys.exit(1)



#!/usr/bin/env python

import asyncio
import copy
import random
import re
import time

import aiohttp
import uvloop

from nettacker.core.lib.base import BaseEngine
from nettacker.core.utils.common import (
    get_http_header_key,
    get_http_header_value,
    replace_dependent_response,
    reverse_and_regex_condition,
)

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


async def perform_request_action(action, request_options):
    start_time = time.time()
    async with action(**request_options) as response:
        return {
            "reason": response.reason,
            "url": str(response.url),
            "status_code": str(response.status),
            "content": await response.content.read(),
            "headers": dict(response.headers),
            "responsetime": time.time() - start_time,
        }


async def send_request(request_options, method):
    async with aiohttp.ClientSession() as session:
        action = getattr(session, method, None)
        response = await asyncio.gather(
            *[asyncio.ensure_future(perform_request_action(action, request_options))]
        )
        return response[0]


def response_conditions_matched(sub_step, response):
    if not response:
        return {}
    condition_type = sub_step["response"]["condition_type"]
    conditions = sub_step["response"]["conditions"]
    condition_results = {}
    for condition in conditions:
        if condition in ["reason", "status_code", "content", "url"]:
            regex = re.findall(re.compile(conditions[condition]["regex"]), response[condition])
            reverse = conditions[condition]["reverse"]
            condition_results[condition] = reverse_and_regex_condition(regex, reverse)
        if condition == "headers":
            # convert headers to case insensitive dict
            for key in response["headers"].copy():
                response["headers"][key.lower()] = response["headers"][key]
            condition_results["headers"] = {}
            for header in conditions["headers"]:
                reverse = conditions["headers"][header]["reverse"]
                try:
                    regex = re.findall(
                        re.compile(conditions["headers"][header]["regex"]),
                        response["headers"][header.lower()]
                        if header.lower() in response["headers"]
                        else False,
                    )
                    condition_results["headers"][header] = reverse_and_regex_condition(
                        regex, reverse
                    )
                except TypeError:
                    condition_results["headers"][header] = []
        if condition == "responsetime":
            if len(conditions[condition].split()) == 2 and conditions[condition].split()[0] in [
                "==",
                "!=",
                ">=",
                "<=",
                ">",
                "<",
            ]:
                exec(
                    "condition_results['responsetime'] = response['responsetime'] if ("
                    + "response['responsetime'] {0} float(conditions['responsetime'].split()[-1])".format(
                        conditions["responsetime"].split()[0]
                    )
                    + ") else []"
                )
            else:
                condition_results["responsetime"] = []
    if condition_type.lower() == "or":
        # if one of the values are matched, it will be a string or float object in the array
        # we count False in the array and if it's not all []; then we know one of the conditions
        # is matched.
        if (
            "headers" not in condition_results
            and (
                list(condition_results.values()).count([]) != len(list(condition_results.values()))
            )
        ) or (
            "headers" in condition_results
            and (
                len(list(condition_results.values()))
                + len(list(condition_results["headers"].values()))
                - list(condition_results.values()).count([])
                - list(condition_results["headers"].values()).count([])
                - 1
                != 0
            )
        ):
            if sub_step["response"].get("log", False):
                condition_results["log"] = sub_step["response"]["log"]
                if "response_dependent" in condition_results["log"]:
                    condition_results["log"] = replace_dependent_response(
                        condition_results["log"], condition_results
                    )
            return condition_results
        else:
            return {}
    if condition_type.lower() == "and":
        if [] in condition_results.values() or (
            "headers" in condition_results and [] in condition_results["headers"].values()
        ):
            return {}
        else:
            if sub_step["response"].get("log", False):
                condition_results["log"] = sub_step["response"]["log"]
                if "response_dependent" in condition_results["log"]:
                    condition_results["log"] = replace_dependent_response(
                        condition_results["log"], condition_results
                    )
            return condition_results
    return {}


class HttpEngine(BaseEngine):
    def run(
        self,
        sub_step,
        module_name,
        target,
        scan_id,
        options,
        process_number,
        module_thread_number,
        total_module_thread_number,
        request_number_counter,
        total_number_of_requests,
    ):
        if options["http_header"] is not None:
            for header in options["http_header"]:
                key = get_http_header_key(header).strip()
                value = get_http_header_value(header)
                if value is not None:
                    sub_step["headers"][key] = value.strip()
                else:
                    sub_step["headers"][key] = ""
        backup_method = copy.deepcopy(sub_step["method"])
        backup_response = copy.deepcopy(sub_step["response"])
        backup_iterative_response_match = copy.deepcopy(
            sub_step["response"]["conditions"].get("iterative_response_match", None)
        )
        if options["user_agent"] == "random_user_agent":
            sub_step["headers"]["User-Agent"] = random.choice(options["user_agents"])
        del sub_step["method"]
        if "dependent_on_temp_event" in backup_response:
            temp_event = self.get_dependent_results_from_database(
                target,
                module_name,
                scan_id,
                backup_response["dependent_on_temp_event"],
            )
            sub_step = self.replace_dependent_values(sub_step, temp_event)
        backup_response = copy.deepcopy(sub_step["response"])
        del sub_step["response"]
        for _i in range(options["retries"]):
            try:
                response = asyncio.run(send_request(sub_step, backup_method))
                response["content"] = response["content"].decode(errors="ignore")
                break
            except Exception:
                response = []
        sub_step["method"] = backup_method
        sub_step["response"] = backup_response

        if backup_iterative_response_match is not None:
            backup_iterative_response_match = copy.deepcopy(
                sub_step["response"]["conditions"].get("iterative_response_match")
            )
            del sub_step["response"]["conditions"]["iterative_response_match"]

        sub_step["response"]["conditions_results"] = response_conditions_matched(
            sub_step, response
        )

        if backup_iterative_response_match is not None and (
            sub_step["response"]["conditions_results"]
            or sub_step["response"]["condition_type"] == "or"
        ):
            sub_step["response"]["conditions"][
                "iterative_response_match"
            ] = backup_iterative_response_match
            for key in sub_step["response"]["conditions"]["iterative_response_match"]:
                result = response_conditions_matched(
                    sub_step["response"]["conditions"]["iterative_response_match"][key],
                    response,
                )
                if result:
                    sub_step["response"]["conditions_results"][key] = result

        return self.process_conditions(
            sub_step,
            module_name,
            target,
            scan_id,
            options,
            response,
            process_number,
            module_thread_number,
            total_module_thread_number,
            request_number_counter,
            total_number_of_requests,
        )



import copy
import json
import re
import time
from abc import ABC
from datetime import datetime

import yaml

from nettacker.config import Config
from nettacker.core.messages import messages as _
from nettacker.core.utils.common import merge_logs_to_list, remove_sensitive_header_keys
from nettacker.database.db import find_temp_events, submit_logs_to_db, submit_temp_logs_to_db
from nettacker.logger import TerminalCodes, get_logger

log = get_logger()


class BaseLibrary(ABC):
    """Nettacker library base class."""

    client = None

    def brute_force(self):
        """Brute force method."""


class BaseEngine(ABC):
    """Nettacker engine base class."""

    library = None

    def apply_extra_data(self, *args, **kwargs):
        """Add extra data into step context."""

    def filter_large_content(self, content, filter_rate=150):
        if len(content) <= filter_rate:
            return content
        else:
            filter_rate -= 1
            filter_index = filter_rate
            for char in content[filter_rate:]:
                if char == " ":
                    return content[0:filter_index] + _("filtered_content")
                else:
                    filter_index += 1
            return content

    def get_dependent_results_from_database(self, target, module_name, scan_id, event_names):
        events = []
        for event_name in event_names.split(","):
            while True:
                event = find_temp_events(target, module_name, scan_id, event_name)
                if event:
                    events.append(json.loads(event)["response"]["conditions_results"])
                    break
                time.sleep(0.1)
        return events

    def find_and_replace_dependent_values(self, sub_step, dependent_on_temp_event):
        if isinstance(sub_step, dict):
            for key in copy.deepcopy(sub_step):
                if not isinstance(sub_step[key], (str, float, int, bytes)):
                    sub_step[key] = self.find_and_replace_dependent_values(
                        copy.deepcopy(sub_step[key]), dependent_on_temp_event
                    )
                else:
                    if isinstance(sub_step[key], str):
                        if "dependent_on_temp_event" in sub_step[key]:
                            globals().update(locals())
                            generate_new_step = copy.deepcopy(sub_step[key])
                            key_name = re.findall(
                                re.compile(
                                    "dependent_on_temp_event\\[\\S+\\]\\['\\S+\\]\\[\\S+\\]"
                                ),
                                generate_new_step,
                            )[0]
                            try:
                                key_value = eval(key_name)
                            except Exception:
                                key_value = "error"
                            sub_step[key] = sub_step[key].replace(key_name, key_value)
        if isinstance(sub_step, list):
            value_index = 0
            for key in copy.deepcopy(sub_step):
                if type(sub_step[value_index]) not in [str, float, int, bytes]:
                    sub_step[key] = self.find_and_replace_dependent_values(
                        copy.deepcopy(sub_step[value_index]), dependent_on_temp_event
                    )
                else:
                    if isinstance(sub_step[value_index], str):
                        if "dependent_on_temp_event" in sub_step[value_index]:
                            globals().update(locals())
                            generate_new_step = copy.deepcopy(sub_step[key])
                            key_name = re.findall(
                                re.compile("dependent_on_temp_event\\['\\S+\\]\\[\\S+\\]"),
                                generate_new_step,
                            )[0]
                            try:
                                key_value = eval(key_name)
                            except Exception:
                                key_value = "error"
                            sub_step[value_index] = sub_step[value_index].replace(
                                key_name, key_value
                            )
                value_index += 1
        return sub_step

    def process_conditions(
        self,
        event,
        module_name,
        target,
        scan_id,
        options,
        response,
        process_number,
        module_thread_number,
        total_module_thread_number,
        request_number_counter,
        total_number_of_requests,
    ):
        # Remove sensitive keys from headers before submitting to DB
        event = remove_sensitive_header_keys(event)
        if "save_to_temp_events_only" in event.get("response", ""):
            submit_temp_logs_to_db(
                {
                    "date": datetime.now(),
                    "target": target,
                    "module_name": module_name,
                    "scan_id": scan_id,
                    "event_name": event["response"]["save_to_temp_events_only"],
                    "port": event.get("ports", ""),
                    "event": event,
                    "data": response,
                }
            )
        if event["response"]["conditions_results"] and "save_to_temp_events_only" not in event.get(
            "response", ""
        ):
            # remove sensitive information before submitting to db

            options = copy.deepcopy(options)
            for key in Config.api:
                try:
                    del options[key]
                except KeyError:
                    continue

            del event["response"]["conditions"]
            del event["response"]["condition_type"]
            if "log" in event["response"]:
                del event["response"]["log"]
            event_request_keys = copy.deepcopy(event)
            del event_request_keys["response"]
            submit_logs_to_db(
                {
                    "date": datetime.now(),
                    "target": target,
                    "module_name": module_name,
                    "scan_id": scan_id,
                    "port": event.get("ports")
                    or event.get("port")
                    or (
                        event.get("url").split(":")[2].split("/")[0]
                        if isinstance(event.get("url"), str)
                        and len(event.get("url").split(":")) >= 3
                        and event.get("url").split(":")[2].split("/")[0].isdigit()
                        else ""
                    ),
                    "event": " ".join(yaml.dump(event_request_keys).split())
                    + " conditions: "
                    + " ".join(yaml.dump(event["response"]["conditions_results"]).split()),
                    "json_event": event,
                }
            )
            log_list = merge_logs_to_list(event["response"]["conditions_results"])
            if log_list:
                log.success_event_info(
                    _("send_success_event_from_module").format(
                        process_number,
                        module_name,
                        target,
                        module_thread_number,
                        total_module_thread_number,
                        request_number_counter,
                        total_number_of_requests,
                        " ",
                        self.filter_large_content(
                            "\n".join(
                                [
                                    TerminalCodes.PURPLE.value + key + TerminalCodes.RESET.value
                                    for key in log_list
                                ]
                            ),
                            filter_rate=100000,
                        ),
                    )
                )
            else:
                log.success_event_info(
                    _("send_success_event_from_module").format(
                        process_number,
                        module_name,
                        target,
                        module_thread_number,
                        total_module_thread_number,
                        request_number_counter,
                        total_number_of_requests,
                        " ".join(
                            [
                                TerminalCodes.YELLOW.value + key + TerminalCodes.RESET.value
                                if ":" in key
                                else TerminalCodes.GREEN.value + key + TerminalCodes.RESET.value
                                for key in yaml.dump(event_request_keys).split()
                            ]
                        ),
                        self.filter_large_content(
                            "conditions: "
                            + " ".join(
                                [
                                    TerminalCodes.PURPLE.value + key + TerminalCodes.RESET.value
                                    if ":" in key
                                    else TerminalCodes.GREEN.value
                                    + key
                                    + TerminalCodes.RESET.value
                                    for key in yaml.dump(
                                        event["response"]["conditions_results"]
                                    ).split()
                                ]
                            ),
                            filter_rate=150,
                        ),
                    )
                )
            log.verbose_info(json.dumps(event))
            return True
        else:
            del event["response"]["conditions"]
            log.verbose_info(
                _("send_unsuccess_event_from_module").format(
                    process_number,
                    module_name,
                    target,
                    module_thread_number,
                    total_module_thread_number,
                    request_number_counter,
                    total_number_of_requests,
                )
            )
            log.verbose_info(json.dumps(event))
            return "save_to_temp_events_only" in event["response"]

    def replace_dependent_values(self, sub_step, dependent_on_temp_event):
        return self.find_and_replace_dependent_values(sub_step, dependent_on_temp_event)

    def run(
        self,
        sub_step,
        module_name,
        target,
        scan_id,
        options,
        process_number,
        module_thread_number,
        total_module_thread_number,
        request_number_counter,
        total_number_of_requests,
    ):
        """Engine entry point."""
        backup_method = copy.deepcopy(sub_step["method"])
        backup_response = copy.deepcopy(sub_step["response"])
        del sub_step["method"]
        del sub_step["response"]

        for attr_name in ("ports", "usernames", "passwords"):
            if attr_name in sub_step:
                value = sub_step.pop(attr_name)
                sub_step[attr_name.rstrip("s")] = int(value) if attr_name == "ports" else value

        if "dependent_on_temp_event" in backup_response:
            temp_event = self.get_dependent_results_from_database(
                target, module_name, scan_id, backup_response["dependent_on_temp_event"]
            )
            sub_step = self.replace_dependent_values(sub_step, temp_event)

        action = getattr(self.library(), backup_method)
        for _i in range(options["retries"]):
            try:
                response = action(**sub_step)
                break
            except Exception:
                response = []

        sub_step["method"] = backup_method
        sub_step["response"] = backup_response
        sub_step["response"]["conditions_results"] = response

        self.apply_extra_data(sub_step, response)

        return self.process_conditions(
            sub_step,
            module_name,
            target,
            scan_id,
            options,
            response,
            process_number,
            module_thread_number,
            total_module_thread_number,
            request_number_counter,
            total_number_of_requests,
        )



import ftplib

from nettacker.core.lib.base import BaseEngine, BaseLibrary


class FtpLibrary(BaseLibrary):
    client = ftplib.FTP

    def brute_force(self, host, port, username, password, timeout):
        connection = self.client(timeout=timeout)
        try:
            connection.connect(host, port)
            connection.login(username, password)
        finally:
            connection.close()

        return {
            "host": host,
            "port": port,
            "username": username,
            "password": password,
        }


class FtpEngine(BaseEngine):
    library = FtpLibrary



import json
import time

try:
    import apsw
except ImportError:
    apsw = None

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from nettacker import logger
from nettacker.api.helpers import structure
from nettacker.config import Config
from nettacker.core.messages import messages
from nettacker.database.models import HostsLog, Report, TempEvents

config = Config()
logger = logger.get_logger()


def db_inputs(connection_type):
    """
    a function to determine the type of database the user wants to work with and
    selects the corresponding connection to the db

    Args:
        connection_type: type of db we are working with

    Returns:
        corresponding command to connect to the db
    """
    context = Config.db.as_dict()
    return {
        "postgres": "postgresql+psycopg2://{username}:{password}@{host}:{port}/{name}?sslmode={ssl_mode}".format(
            **context
        ),
        "mysql": "mysql+pymysql://{username}:{password}@{host}:{port}/{name}".format(**context),
        "sqlite": "sqlite:///{name}".format(**context),
    }[connection_type]


def create_connection():
    """
    a function to create connections to db with pessimistic approach

    For sqlite, it creates and returns a sqlite connection object
    for mysql and postgresql, it returns the connection or False if
    connection failed.
    """
    if Config.db.engine.startswith("sqlite") and Config.settings.use_apsw_for_sqlite:
        if apsw is None:
            raise ImportError(messages("apsw_ill_configuration"))

        # In case of sqlite, the name parameter is the database path.
        try:
            DB_PATH = config.db.as_dict()["name"]
            connection = apsw.Connection(DB_PATH)
            connection.setbusytimeout(int(config.settings.timeout) * 100)
            cursor = connection.cursor()

            # Performance enhancing configurations. Put WAL cause that helps with concurrency.
            cursor.execute(f"PRAGMA journal_mode={Config.db.journal_mode}")
            cursor.execute(f"PRAGMA synchronous={Config.db.synchronous_mode}")

            return connection, cursor
        except Exception as e:
            logger.error(messages("apsw_connection_error").format(e))
            raise

    else:
        connection_args = {}

        if Config.db.engine.startswith("sqlite"):
            connection_args["check_same_thread"] = False

        db_engine = create_engine(
            db_inputs(Config.db.engine),
            connect_args=connection_args,
            pool_size=50,
            pool_pre_ping=True,
        )

        return sessionmaker(bind=db_engine)()


def send_submit_query(session):
    """
    a function to send submit based queries to db
    (such as insert and update or delete), it retries 100 times if
    connection returned an error.

    Args:
        session: session to commit

    Returns:
        True if submitted success otherwise False
    """
    if isinstance(session, tuple):
        connection, _ = session
        try:
            for _ in range(1, Config.settings.max_submit_query_retry):
                try:
                    connection.execute("COMMIT")
                    return True
                except Exception:
                    connection.execute("ROLLBACK")
                    time.sleep(Config.settings.retry_delay)
        finally:
            connection.close()
        # If not already returned then connection failed
        logger.warn(messages("database_connect_fail"))
        return False
    else:
        try:
            for _ in range(1, Config.settings.max_submit_query_retry):
                try:
                    session.commit()
                    return True
                except Exception:
                    session.rollback()
                    time.sleep(Config.settings.retry_delay)
            logger.warn(messages("database_connect_fail"))
            return False
        except Exception:
            logger.warn(messages("database_connect_fail"))
            return False


def submit_report_to_db(event):
    """
    this function created to submit the generated reports into db, the
    files are not stored in db, just the path!

    Args:
        event: event log

    Returns:
        return True if submitted otherwise False
    """
    logger.verbose_info(messages("inserting_report_db"))
    session = create_connection()

    if isinstance(session, tuple):
        connection, cursor = session

        try:
            cursor.execute("BEGIN")
            cursor.execute(
                """
                INSERT INTO reports (date, scan_unique_id, report_path_filename, options)
                VALUES (?, ?, ?, ?)
                """,
                (
                    str(event["date"]),
                    event["scan_id"],
                    event["options"]["report_path_filename"],
                    json.dumps(event["options"]),
                ),
            )
            return send_submit_query(session)
        except Exception:
            cursor.execute("ROLLBACK")
            logger.warn(messages("report_insertion_fail"))
            return False
    else:
        session.add(
            Report(
                date=event["date"],
                scan_unique_id=event["scan_id"],
                report_path_filename=event["options"]["report_path_filename"],
                options=json.dumps(event["options"]),
            )
        )
        return send_submit_query(session)


def remove_old_logs(options):
    """
    this function remove old events (and duplicated)
    from nettacker.database based on target, module, scan_id

    Args:
        options: identifiers

    Returns:
        True if success otherwise False
    """
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session

        try:
            cursor.execute("BEGIN")
            cursor.execute(
                """
                DELETE FROM scan_events
                    WHERE target = ?
                      AND module_name = ?
                      AND scan_unique_id != ?
                      AND scan_unique_id != ?
                """,
                (
                    options["target"],
                    options["module_name"],
                    options["scan_id"],
                    options["scan_compare_id"],
                ),
            )
            return send_submit_query(session)
        except Exception:
            cursor.execute("ROLLBACK")
            logger.warn(messages("remove_logs_fail"))
            return False
        finally:
            cursor.close()
            connection.close()
    else:
        session.query(HostsLog).filter(
            HostsLog.target == options["target"],
            HostsLog.module_name == options["module_name"],
            HostsLog.scan_unique_id != options["scan_id"],
            HostsLog.scan_unique_id != options["scan_compare_id"],
            # Don't remove old logs if they are to be used for the scan reports
        ).delete(synchronize_session=False)
        return send_submit_query(session)


def submit_logs_to_db(log):
    """
    this function created to submit new events into database.
    This requires a little more robust handling in case of
    APSW in order to avoid database lock issues.

    Args:
        log: log event in JSON type

    Returns:
        True if success otherwise False
    """
    if isinstance(log, dict):
        session = create_connection()
        if isinstance(session, tuple):
            connection, cursor = session
            try:
                for _ in range(Config.settings.max_retries):
                    try:
                        if not connection.in_transaction:
                            connection.execute("BEGIN")
                        cursor.execute(
                            """
                            INSERT INTO scan_events (target, date, module_name, scan_unique_id, port, event, json_event)
                            VALUES (?, ?, ?, ?, ?, ?, ?)
                            """,
                            (
                                log["target"],
                                str(log["date"]),
                                log["module_name"],
                                log["scan_id"],
                                json.dumps(log["port"]),
                                json.dumps(log["event"]),
                                json.dumps(log["json_event"]),
                            ),
                        )
                        return send_submit_query(session)

                    except apsw.BusyError as e:
                        if "database is locked" in str(e).lower():
                            logger.warn(
                                messages("database_lock_issue").format(
                                    f"{_ + 1}/{Config.settings.max_retries}"
                                )
                            )
                            if connection.in_transaction:
                                connection.execute("ROLLBACK")
                            time.sleep(Config.settings.retry_delay)
                            continue
                        else:
                            if connection.in_transaction:
                                connection.execute("ROLLBACK")
                            return False
                    except Exception:
                        try:
                            if connection.in_transaction:
                                connection.execute("ROLLBACK")
                        except Exception:
                            pass
                        return False
                # All retries exhausted but we want to continue operation
                logger.warn(messages("database_retries_exhausted"))
                return True
            finally:
                cursor.close()
                connection.close()

        else:
            session.add(
                HostsLog(
                    target=log["target"],
                    date=log["date"],
                    module_name=log["module_name"],
                    scan_unique_id=log["scan_id"],
                    port=json.dumps(log["port"]),
                    event=json.dumps(log["event"]),
                    json_event=json.dumps(log["json_event"]),
                )
            )
            return send_submit_query(session)
    else:
        logger.warn(messages("invalid_json_type_to_db").format(log))
        return False


def submit_temp_logs_to_db(log):
    """
    this function created to submit new events into database.
    This requires a little more robust handling in case of
    APSW in order to avoid database lock issues.

    Args:
        log: log event in JSON type

    Returns:
        True if success otherwise False
    """
    if isinstance(log, dict):
        session = create_connection()
        if isinstance(session, tuple):
            connection, cursor = session

            try:
                for _ in range(Config.settings.max_retries):
                    try:
                        if not connection.in_transaction:
                            cursor.execute("BEGIN")
                        cursor.execute(
                            """
                            INSERT INTO temp_events (target, date, module_name, scan_unique_id, event_name, port, event, data)
                            VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                            """,
                            (
                                log["target"],
                                str(log["date"]),
                                log["module_name"],
                                log["scan_id"],
                                log["event_name"],
                                json.dumps(log["port"]),
                                json.dumps(log["event"]),
                                json.dumps(log["data"]),
                            ),
                        )
                        return send_submit_query(session)
                    except apsw.BusyError as e:
                        if "database is locked" in str(e).lower():
                            logger.warn(
                                messages("database_lock_issue").format(
                                    f"{_ + 1}/{Config.settings.max_retries}"
                                )
                            )
                            try:
                                if connection.in_transaction:
                                    connection.execute("ROLLBACK")
                            except Exception:
                                pass
                            time.sleep(Config.settings.retry_delay)
                            continue
                        else:
                            try:
                                if connection.in_transaction:
                                    connection.execute("ROLLBACK")
                            except Exception:
                                pass
                            return False
                    except Exception:
                        try:
                            if connection.in_transaction:
                                connection.execute("ROLLBACK")
                        except Exception:
                            pass
                        return False
                # All retries exhausted but we want to continue operation
                logger.warn(messages("database_retries_exhausted"))
                return True
            finally:
                cursor.close()
                connection.close()
        else:
            session.add(
                TempEvents(
                    target=log["target"],
                    date=log["date"],
                    module_name=log["module_name"],
                    scan_unique_id=log["scan_id"],
                    event_name=log["event_name"],
                    port=json.dumps(log["port"]),
                    event=json.dumps(log["event"]),
                    data=json.dumps(log["data"]),
                )
            )
            return send_submit_query(session)
    else:
        logger.warn(messages("invalid_json_type_to_db").format(log))
        return False


def find_temp_events(target, module_name, scan_id, event_name):
    """
    select all events by scan_unique id, target, module_name

    Args:
        target: target
        module_name: module name
        scan_id: unique scan identifier
        event_name: event_name

    Returns:
        an array with JSON events or an empty array
    """
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session
        try:
            cursor.execute(
                """
                SELECT event
                FROM temp_events
                WHERE target = ? AND module_name = ? AND scan_unique_id = ? AND event_name = ?
                LIMIT 1
            """,
                (target, module_name, scan_id, event_name),
            )

            row = cursor.fetchone()
            if row:
                return row[0]
            return []
        except Exception:
            logger.warn(messages("database_connect_fail"))
            return []
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass
    else:
        result = (
            session.query(TempEvents)
            .filter(
                TempEvents.target == target,
                TempEvents.module_name == module_name,
                TempEvents.scan_unique_id == scan_id,
                TempEvents.event_name == event_name,
            )
            .first()
        )

        return result.event if result else []


def find_events(target, module_name, scan_id):
    """
    select all events by scan_unique id, target, module_name

    Args:
        target: target
        module_name: module name
        scan_id: unique scan identifier

    Returns:
        an array with JSON events or an empty array
    """
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session

        try:
            cursor.execute(
                """
                SELECT json_event FROM scan_events
                WHERE target = ? AND module_name = ? and scan_unique_id = ?
                """,
                (target, module_name, scan_id),
            )

            rows = cursor.fetchall()
            if rows:
                return [json.dumps((json.loads(row[0]))) for row in rows]
            return []
        except Exception:
            logger.warn(messages("database_query_fail"))
            return []
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass
    else:
        return [
            row.json_event
            for row in session.query(HostsLog)
            .filter(
                HostsLog.target == target,
                HostsLog.module_name == module_name,
                HostsLog.scan_unique_id == scan_id,
            )
            .all()
        ]


def select_reports(page):
    """
    this function created to crawl into submitted results,
    it shows last 10 results submitted in the database.
    you may change the page (default 1) to go to next/previous page.

    Args:
        page: page number

    Returns:
        list of events in array and JSON type, otherwise an error in JSON type.
    """
    selected = []
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session
        offset = (page - 1) * 10

        try:
            cursor.execute(
                """
                SELECT id, date, scan_unique_id, report_path_filename, options
                FROM reports
                ORDER BY id DESC
                LIMIT 10 OFFSET ?
                """,
                (offset,),
            )

            rows = cursor.fetchall()

            for row in rows:
                tmp = {
                    "id": row[0],
                    "date": str(row[1]),
                    "scan_id": row[2],
                    "report_path_filename": row[3],
                    "options": json.loads(row[4]),
                }
                selected.append(tmp)
            return selected

        except Exception:
            logger.warn(messages("report_retrieval_fail"))
            return structure(status="error", msg=messages("database_error"))
        finally:
            cursor.close()
            connection.close()
    else:
        try:
            search_data = (
                session.query(Report).order_by(Report.id.desc()).offset((page * 10) - 10).limit(10)
            )
            for data in search_data:
                tmp = {
                    "id": data.id,
                    "date": data.date,
                    "scan_id": data.scan_unique_id,
                    "report_path_filename": data.report_path_filename,
                    "options": json.loads(data.options),
                }
                selected.append(tmp)
        except Exception:
            return structure(status="error", msg=messages("database_error"))
        return selected


def get_scan_result(id):
    """
    this function created to download results by the result ID.

    Args:
        id: scan id

    Returns:
        result file content (TEXT, HTML, JSON) if success otherwise and error in JSON type.
    """
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session
        try:
            cursor.execute(
                "SELECT report_path_filename from reports WHERE id = ?",
                (id,),
            )

            row = cursor.fetchone()
            if row:
                filename = row[0]
                try:
                    with open(str(filename), "rb") as fp:
                        contents = fp.read()
                    return filename, contents
                except IOError as e:
                    logger.error(messages("read_report_fail").format(e))
                    return None
            else:
                return structure(status="error", msg=messages("database_error"))
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass
    else:
        report = session.query(Report).filter_by(id=id).first()
        if not report:
            return None

        try:
            with open(str(report.report_path_filename), "rb") as fp:
                contents = fp.read()
            return report.report_path_filename, contents
        except IOError as e:
            logger.error(messages("read_report_fail").format(e))
            return None


def last_host_logs(page):
    """
    this function created to select the last 10 events from the database.
    you can goto next page by changing page value.

    Args:
        page: page number

    Returns:
        an array of events in JSON type if success otherwise an error in JSON type
    """
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session
        try:
            cursor.execute(
                """
                SELECT DISTINCT target 
                FROM scan_events
                ORDER BY id DESC 
                LIMIT 10 OFFSET ?
                """,
                [(page - 1) * 10],
            )
            targets = cursor.fetchall()

            if not targets:
                cursor.close()
                connection.close()
                return structure(status="finished", msg=messages("search_results_end"))

            hosts = []

            for (target,) in targets:
                cursor.execute(
                    """
                    SELECT DISTINCT module_name 
                    FROM scan_events
                    WHERE target = ?
                    """,
                    [target],
                )
                module_names = [row[0] for row in cursor.fetchall()]

                cursor.execute(
                    """
                    SELECT date 
                    FROM scan_events
                    WHERE target = ? 
                    ORDER BY id DESC 
                    LIMIT 1
                    """,
                    [target],
                )
                latest_date = cursor.fetchone()
                latest_date = latest_date[0] if latest_date else None

                cursor.execute(
                    """
                    SELECT event 
                    FROM scan_events
                    WHERE target = ?
                    """,
                    [target],
                )
                events = [row[0] for row in cursor.fetchall()]

                hosts.append(
                    {
                        "target": target,
                        "info": {
                            "module_name": module_names,
                            "date": latest_date,
                            "events": events,
                        },
                    }
                )
            return hosts

        except Exception:
            logger.warn(messages("database_query_fail"))
            return structure(status="error", msg=messages("database_error"))
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass
    else:
        hosts = [
            {
                "target": host.target,
                "info": {
                    "module_name": [
                        _.module_name
                        for _ in session.query(HostsLog)
                        .filter(HostsLog.target == host.target)
                        .group_by(HostsLog.module_name)
                        .all()
                    ],
                    "date": session.query(HostsLog)
                    .filter(HostsLog.target == host.target)
                    .order_by(HostsLog.id.desc())
                    .first()
                    .date,
                    "events": [
                        _.event
                        for _ in session.query(HostsLog)
                        .filter(HostsLog.target == host.target)
                        .all()
                    ],
                },
            }
            for host in session.query(HostsLog)
            .group_by(HostsLog.target)
            .order_by(HostsLog.id.desc())
            .offset((page * 10) - 10)
            .limit(10)
        ]
        if not hosts:
            return structure(status="finished", msg=messages("search_results_end"))
        return hosts


def get_logs_by_scan_id(scan_id):
    """
    select all events by scan id hash

    Args:
        scan_id: scan id hash

    Returns:
        an array with JSON events or an empty array
    """
    session = create_connection()

    if isinstance(session, tuple):
        connection, cursor = session
        try:
            cursor.execute(
                "SELECT scan_unique_id, target, module_name, date, port, event, json_event from scan_events WHERE scan_unique_id = ?",
                (scan_id,),  # We have to put this as an indexed element
            )
            rows = cursor.fetchall()
            return [
                {
                    "scan_id": row[0],
                    "target": row[1],
                    "module_name": row[2],
                    "date": str(row[3]),
                    "port": json.loads(row[4]),
                    "event": json.loads(row[5]),
                    "json_event": json.loads(row[6]) if row[6] else {},
                }
                for row in rows
            ]
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass
    else:
        return [
            {
                "scan_id": scan_id,
                "target": log.target,
                "module_name": log.module_name,
                "date": str(log.date),
                "port": json.loads(log.port),
                "event": json.loads(log.event),
                "json_event": log.json_event,
            }
            for log in session.query(HostsLog).filter(HostsLog.scan_unique_id == scan_id).all()
        ]


def get_options_by_scan_id(scan_id):
    """
    select all stored options of the scan by scan id hash
    Args:
        scan_id: scan id hash
    Returns:
        an array with a dict with stored options or an empty array
    """
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session

        try:
            cursor.execute(
                "SELECT options from reports WHERE scan_unique_id = ?",
                (scan_id,),
            )
            rows = cursor.fetchall()
            if rows:
                return [{"options": row[0]} for row in rows]
            return []
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass

    else:
        return [
            {"options": log.options}
            for log in session.query(Report).filter(Report.scan_unique_id == scan_id).all()
        ]


def logs_to_report_json(target):
    """
    select all reports of a host

    Args:
        host: the host to search

    Returns:
        an array with JSON events or an empty array
    """
    try:
        session = create_connection()
        if isinstance(session, tuple):
            connection, cursor = session
            return_logs = []
            try:
                cursor.execute(
                    "SELECT scan_unique_id, target, port, event, json_event FROM scan_events WHERE target = ?",
                    (target,),
                )
                rows = cursor.fetchall()
                if rows:
                    for log in rows:
                        data = {
                            "scan_id": log[0],
                            "target": log[1],
                            "port": json.loads(log[2]),
                            "event": json.loads(log[3]),
                            "json_event": json.loads(log[4]),
                        }
                        return_logs.append(data)
                return return_logs
            finally:
                try:
                    cursor.close()
                    connection.close()
                except Exception:
                    pass
        else:
            return_logs = []
            logs = session.query(HostsLog).filter(HostsLog.target == target)
            for log in logs:
                data = {
                    "scan_id": log.scan_unique_id,
                    "target": log.target,
                    "port": json.loads(log.port),
                    "event": json.loads(log.event),
                    "json_event": json.loads(log.json_event),
                }
                return_logs.append(data)
            return return_logs

    except Exception:
        return []


def logs_to_report_html(target):
    """
    generate HTML report with d3_tree_v2_graph for a host

    Args:
        target: the target

    Returns:
        HTML report
    """
    from nettacker.core.graph import build_graph
    from nettacker.lib.html_log import log_data

    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session
        try:
            cursor.execute(
                """
                SELECT date, target, module_name, scan_unique_id, port, event, json_event
                FROM scan_events
                WHERE target = ?
                """,
                (target,),
            )

            rows = cursor.fetchall()
            logs = [
                {
                    "date": log[0],
                    "target": log[1],
                    "module_name": log[2],
                    "scan_id": log[3],
                    "port": log[4],
                    "event": log[5],
                    "json_event": log[6],
                }
                for log in rows
            ]
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass

        html_graph = build_graph("d3_tree_v2_graph", logs)

        html_content = log_data.table_title.format(
            html_graph,
            log_data.css_1,
            "date",
            "target",
            "module_name",
            "scan_id",
            "port",
            "event",
            "json_event",
        )
        for event in logs:
            html_content += log_data.table_items.format(
                event["date"],
                event["target"],
                event["module_name"],
                event["scan_id"],
                event["port"],
                event["event"],
                event["json_event"],
            )
        html_content += (
            log_data.table_end + '<p class="footer">' + messages("nettacker_report") + "</p>"
        )
        return html_content
    else:
        logs = [
            {
                "date": log.date,
                "target": log.target,
                "module_name": log.module_name,
                "scan_id": log.scan_unique_id,
                "port": log.port,
                "event": log.event,
                "json_event": log.json_event,
            }
            for log in session.query(HostsLog).filter(HostsLog.target == target).all()
        ]
        html_graph = build_graph("d3_tree_v2_graph", logs)

        html_content = log_data.table_title.format(
            html_graph,
            log_data.css_1,
            "date",
            "target",
            "module_name",
            "scan_id",
            "port",
            "event",
            "json_event",
        )
        for event in logs:
            html_content += log_data.table_items.format(
                event["date"],
                event["target"],
                event["module_name"],
                event["scan_id"],
                event["port"],
                event["event"],
                event["json_event"],
            )
        html_content += (
            log_data.table_end + '<p class="footer">' + messages("nettacker_report") + "</p>"
        )
        return html_content


def search_logs(page, query):
    """
    search in events (host, date, port, module, category, description,
    username, password, scan_id, scan_cmd)

    Args:
        page: page number
        query: query to search

    Returns:
        an array with JSON structure of founded events or an empty array
    """
    selected = []
    session = create_connection()
    if isinstance(session, tuple):
        connection, cursor = session
        try:
            # Fetch targets matching the query
            cursor.execute(
                """
                SELECT DISTINCT target FROM scan_events
                WHERE target LIKE ? OR date LIKE ? OR module_name LIKE ?
                OR port LIKE ? OR event LIKE ? OR scan_unique_id LIKE ?
                ORDER BY id DESC
                LIMIT 10 OFFSET ?
                """,
                (
                    f"%{query}%",
                    f"%{query}%",
                    f"%{query}%",
                    f"%{query}%",
                    f"%{query}%",
                    f"%{query}%",
                    (page * 10) - 10,
                ),
            )
            targets = cursor.fetchall()
            for target_row in targets:
                target = target_row[0]
                # Fetch data for each target grouped by key fields
                cursor.execute(
                    """
                    SELECT date, module_name, port, event, json_event FROM scan_events
                    WHERE target = ?
                    GROUP BY module_name, port, scan_unique_id, event
                    ORDER BY id DESC
                    """,
                    (target,),
                )
                results = cursor.fetchall()

                tmp = {
                    "target": target,
                    "info": {
                        "module_name": [],
                        "port": [],
                        "date": [],
                        "event": [],
                        "json_event": [],
                    },
                }

                for data in results:
                    date, module_name, port, event, json_event = data
                    if module_name not in tmp["info"]["module_name"]:
                        tmp["info"]["module_name"].append(module_name)
                    if date not in tmp["info"]["date"]:
                        tmp["info"]["date"].append(date)
                    parsed_port = json.loads(port)
                    if parsed_port not in tmp["info"]["port"]:
                        tmp["info"]["port"].append(parsed_port)
                    parsed_event = json.loads(event)
                    if parsed_event not in tmp["info"]["event"]:
                        tmp["info"]["event"].append(parsed_event)
                    parsed_json_event = json.loads(json_event)
                    if parsed_json_event not in tmp["info"]["json_event"]:
                        tmp["info"]["json_event"].append(parsed_json_event)

                selected.append(tmp)

        except Exception:
            return structure(status="error", msg=messages("database_error"))
        finally:
            try:
                cursor.close()
                connection.close()
            except Exception:
                pass
        if not selected:
            return structure(status="finished", msg=messages("search_results_end"))
        return selected
    else:
        try:
            for host in (
                session.query(HostsLog)
                .filter(
                    (HostsLog.target.like("%" + str(query) + "%"))
                    | (HostsLog.date.like("%" + str(query) + "%"))
                    | (HostsLog.module_name.like("%" + str(query) + "%"))
                    | (HostsLog.port.like("%" + str(query) + "%"))
                    | (HostsLog.event.like("%" + str(query) + "%"))
                    | (HostsLog.scan_unique_id.like("%" + str(query) + "%"))
                )
                .group_by(HostsLog.target)
                .order_by(HostsLog.id.desc())
                .offset((page * 10) - 10)
                .limit(10)
            ):
                for data in (
                    session.query(HostsLog)
                    .filter(HostsLog.target == str(host.target))
                    .group_by(
                        HostsLog.module_name,
                        HostsLog.port,
                        HostsLog.scan_unique_id,
                        HostsLog.event,
                    )
                    .order_by(HostsLog.id.desc())
                    .all()
                ):
                    n = 0
                    capture = None
                    for selected_data in selected:
                        if selected_data["target"] == host.target:
                            capture = n
                        n += 1
                    if capture is None:
                        tmp = {
                            "target": data.target,
                            "info": {
                                "module_name": [],
                                "port": [],
                                "date": [],
                                "event": [],
                                "json_event": [],
                            },
                        }
                        selected.append(tmp)
                        n = 0
                        for selected_data in selected:
                            if selected_data["target"] == host.target:
                                capture = n
                            n += 1
                    if data.target == selected[capture]["target"]:
                        if data.module_name not in selected[capture]["info"]["module_name"]:
                            selected[capture]["info"]["module_name"].append(data.module_name)
                        if data.date not in selected[capture]["info"]["date"]:
                            selected[capture]["info"]["date"].append(data.date)
                        if data.port not in selected[capture]["info"]["port"]:
                            selected[capture]["info"]["port"].append(json.loads(data.port))
                        if data.event not in selected[capture]["info"]["event"]:
                            selected[capture]["info"]["event"].append(json.loads(data.event))
                        if data.json_event not in selected[capture]["info"]["json_event"]:
                            selected[capture]["info"]["json_event"].append(
                                json.loads(data.json_event)
                            )
        except Exception:
            return structure(status="error", msg=messages("database_error"))
        if len(selected) == 0:
            return structure(status="finished", msg=messages("search_results_end"))
        return selected



import smtplib

from nettacker.core.lib.base import BaseEngine, BaseLibrary


class SmtpLibrary(BaseLibrary):
    client = smtplib.SMTP

    def brute_force(self, host, port, username, password, timeout):
        connection = self.client(host, port, timeout=timeout)
        connection.login(username, password)
        connection.close()

        return {
            "host": host,
            "port": port,
            "username": username,
            "password": password,
        }


class SmtpEngine(BaseEngine):
    library = SmtpLibrary



def structure(status="", msg=""):
    """
    basic JSON message structure

    Args:
        status: status (ok, failed)
        msg: the message content

    Returns:
        a JSON message
    """
    return {"status": status, "msg": msg}



import sys
from enum import Enum
from functools import cached_property

from nettacker.core.utils import time


class TerminalCodes(Enum):
    RESET = "\033[0m"

    # Colors \033[1;
    GREY = "\033[1;30m"
    RED = "\033[1;31m"
    GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    BLUE = "\033[1;34m"
    PURPLE = "\033[1;35m"
    CYAN = "\033[1;36m"
    WHITE = "\033[1;37m"


class Logger:
    """Nettacker logger."""

    @staticmethod
    def log(text):
        print(text, end="", flush=True)  # noqa: T201

    @cached_property
    def run_from_api(self):
        """
        check if framework run from API to prevent any alert

        Returns:
            True if run from API otherwise False
        """
        return "--start-api" in sys.argv

    @cached_property
    def verbose_mode_is_enabled(self):
        return "--verbose" in sys.argv or "-v" in sys.argv

    @cached_property
    def event_verbose_mode_is_enabled(self):
        return "--verbose-event" in sys.argv

    def info(self, content):
        """
        build the info message, log the message in database if requested,
        rewrite the thread temporary file

        Args:
            content: content of the message

        Returns:
            None
        """
        if not self.run_from_api:
            self.log(
                TerminalCodes.YELLOW.value
                + "[{0}][+] ".format(time.now())
                + TerminalCodes.GREEN.value
                + content
                + TerminalCodes.RESET.value
                + "\n",
            )

    def verbose_event_info(self, content):
        """
        build the info message, log the message in database if requested,
        rewrite the thread temporary file

        Args:
            content: content of the message

        Returns:
            None
        """
        if not self.run_from_api and (
            self.verbose_mode_is_enabled or self.event_verbose_mode_is_enabled
        ):  # prevent to stdout if run from API
            self.log(
                TerminalCodes.YELLOW.value
                + "[{0}][+] ".format(time.now())
                + TerminalCodes.GREEN.value
                + content
                + TerminalCodes.RESET.value
                + "\n",
            )

    def write(self, content):
        """
        simple print a message

        Args:
            content: content of the message

        Returns:
            None
        """
        if not self.run_from_api:
            self.log(content)

    def success_event_info(self, content):
        """
        build the info message, log the message in database if requested,
        rewrite the thread temporary file

        Args:
            content: content of the message

        Returns:
            None
        """
        if not self.run_from_api:
            self.log(
                TerminalCodes.RED.value
                + "[{0}][+++] ".format(time.now())
                + TerminalCodes.CYAN.value
                + content
                + TerminalCodes.RESET.value
                + "\n",
            )

    def verbose_info(self, content):
        """
        build the info message, log the message in database if requested,
        rewrite the thread temporary file

        Args:
            content: content of the message

        Returns:
            None
        """
        if self.verbose_mode_is_enabled:
            self.log(
                TerminalCodes.YELLOW.value
                + "[{0}][+] ".format(time.now())
                + TerminalCodes.PURPLE.value
                + content
                + TerminalCodes.RESET.value
                + "\n",
            )

    def warn(self, content):
        """
        build the warn message

        Args:
            content: content of the message

        Returns:
            the message in warn structure - None
        """
        if not self.run_from_api:
            self.log(
                TerminalCodes.BLUE.value
                + "[{0}][!] ".format(time.now())
                + TerminalCodes.YELLOW.value
                + content
                + TerminalCodes.RESET.value
                + "\n",
            )

    def error(self, content):
        """
        build the error message

        Args:
            content: content of the message

        Returns:
            the message in error structure - None
        """
        self.log(
            TerminalCodes.RED.value
            + "[{0}][X] ".format(time.now())
            + TerminalCodes.YELLOW.value
            + content
            + TerminalCodes.RESET.value
            + "\n"
        )

    def write_to_api_console(self, content):
        """
        simple print a message in API mode

        Args:
            content: content of the message

        Returns:
            None
        """
        self.log(content)

    def reset_color(self):
        self.log(TerminalCodes.RESET.value)


def get_logger():
    return Logger()



import json
import sys
from argparse import ArgumentParser

import yaml

from nettacker import all_module_severity_and_desc
from nettacker.config import Config, version_info
from nettacker.core.die import die_failure, die_success
from nettacker.core.ip import (
    generate_ip_range,
    is_ipv4_cidr,
    is_ipv4_range,
    is_ipv6_cidr,
    is_ipv6_range,
    is_single_ipv4,
    is_single_ipv6,
)
from nettacker.core.messages import messages as _
from nettacker.core.template import TemplateLoader
from nettacker.core.utils import common as common_utils
from nettacker.logger import TerminalCodes, get_logger

log = get_logger()


class ArgParser(ArgumentParser):
    def __init__(self, api_arguments=None) -> None:
        super().__init__(prog="Nettacker", add_help=False)

        self.api_arguments = api_arguments
        self.graphs = self.load_graphs()
        self.languages = self.load_languages()

        self.modules = self.load_modules(full_details=True)
        log.info(_("loaded_modules").format(len(self.modules)))

        self.profiles = self.load_profiles()

        self.add_arguments()
        self.parse_arguments()

    @staticmethod
    def load_graphs():
        """
        load all available graphs

        Returns:
            an array of graph names
        """

        graph_names = []
        for graph_library in Config.path.graph_dir.glob("*/engine.py"):
            graph_names.append(str(graph_library).split("/")[-2] + "_graph")
        return list(set(graph_names))

    @staticmethod
    def load_languages():
        """
        Get available languages

        Returns:
            an array of languages
        """
        languages_list = []

        for language in Config.path.locale_dir.glob("*.yaml"):
            languages_list.append(str(language).split("/")[-1].split(".")[0])

        return list(set(languages_list))

    @staticmethod
    def load_modules(limit=-1, full_details=False):
        """
        load all available modules

        limit: return limited number of modules
        full: with full details

        Returns:
            an array of all module names
        """
        # Search for Modules
        module_names = {}
        for module_name in sorted(Config.path.modules_dir.glob("**/*.yaml")):
            library = str(module_name).split("/")[-1].split(".")[0]
            category = str(module_name).split("/")[-2]
            module = f"{library}_{category}"
            contents = yaml.safe_load(TemplateLoader(module).open().split("payload:")[0])
            module_names[module] = contents["info"] if full_details else None
            info = contents.get("info", {})
            all_module_severity_and_desc[module] = {
                "severity": info.get("severity", 0),
                "desc": info.get("description", ""),
            }
            if len(module_names) == limit:
                module_names["..."] = {}
                break
        module_names = common_utils.sort_dictionary(module_names)
        module_names["all"] = {}

        return module_names

    @staticmethod
    def load_profiles(limit=-1):
        """
        load all available profiles

        Returns:
            an array of all profile names
        """
        all_modules_with_details = ArgParser.load_modules(full_details=True).copy()
        profiles = {}
        if "..." in all_modules_with_details:
            del all_modules_with_details["..."]
        del all_modules_with_details["all"]
        for key in all_modules_with_details:
            for tag in all_modules_with_details[key]["profiles"]:
                if tag not in profiles:
                    profiles[tag] = []
                    profiles[tag].append(key)
                else:
                    profiles[tag].append(key)
                if len(profiles) == limit:
                    profiles = common_utils.sort_dictionary(profiles)
                    profiles["..."] = []
                    profiles["all"] = []
                    return profiles
        profiles = common_utils.sort_dictionary(profiles)
        profiles["all"] = []

        return profiles

    def add_arguments(self):
        # Engine Options
        engine_options = self.add_argument_group(_("engine"), _("engine_input"))
        engine_options.add_argument(
            "-L",
            "--language",
            action="store",
            dest="language",
            default=Config.settings.language,
            help=_("select_language").format(self.languages),
        )
        engine_options.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            dest="verbose_mode",
            default=Config.settings.verbose_mode,
            help=_("verbose_mode"),
        )
        engine_options.add_argument(
            "--verbose-event",
            action="store_true",
            dest="verbose_event",
            default=Config.settings.verbose_event,
            help=_("verbose_event"),
        )
        engine_options.add_argument(
            "-V",
            "--version",
            action="store_true",
            default=Config.settings.show_version,
            dest="show_version",
            help=_("software_version"),
        )
        engine_options.add_argument(
            "-o",
            "--output",
            action="store",
            default=Config.settings.report_path_filename,
            dest="report_path_filename",
            help=_("save_logs"),
        )
        engine_options.add_argument(
            "--graph",
            action="store",
            default=Config.settings.graph_name,
            dest="graph_name",
            help=_("available_graph").format(self.graphs),
        )
        engine_options.add_argument(
            "-h",
            "--help",
            action="store_true",
            default=Config.settings.show_help_menu,
            dest="show_help_menu",
            help=_("help_menu"),
        )

        # Target Options
        target_options = self.add_argument_group(_("target"), _("target_input"))
        target_options.add_argument(
            "-i",
            "--targets",
            action="store",
            dest="targets",
            default=Config.settings.targets,
            help=_("target_list"),
        )
        target_options.add_argument(
            "-l",
            "--targets-list",
            action="store",
            dest="targets_list",
            default=Config.settings.targets_list,
            help=_("read_target"),
        )

        # Exclude Module Name
        exclude_modules = sorted(self.modules.keys())[:10]
        exclude_modules.remove("all")

        # Method Options
        method_options = self.add_argument_group(_("Method"), _("scan_method_options"))
        method_options.add_argument(
            "-m",
            "--modules",
            action="store",
            dest="selected_modules",
            default=Config.settings.selected_modules,
            help=_("choose_scan_method").format(list(self.modules.keys())[:10]),
        )
        method_options.add_argument(
            "--modules-extra-args",
            action="store",
            dest="modules_extra_args",
            default=Config.settings.modules_extra_args,
            help=_("modules_extra_args_help"),
        )
        method_options.add_argument(
            "--show-all-modules",
            action="store_true",
            dest="show_all_modules",
            default=Config.settings.show_all_modules,
            help=_("show_all_modules"),
        )
        method_options.add_argument(
            "--profile",
            action="store",
            default=Config.settings.profiles,
            dest="profiles",
            help=_("select_profile").format(list(self.profiles.keys())[:10]),
        )
        method_options.add_argument(
            "--show-all-profiles",
            action="store_true",
            dest="show_all_profiles",
            default=Config.settings.show_all_profiles,
            help=_("show_all_profiles"),
        )
        method_options.add_argument(
            "-x",
            "--exclude-modules",
            action="store",
            dest="excluded_modules",
            default=Config.settings.excluded_modules,
            help=_("exclude_scan_method").format(exclude_modules),
        )
        method_options.add_argument(
            "-X",
            "--exclude-ports",
            action="store",
            dest="excluded_ports",
            default=Config.settings.excluded_ports,
            help=_("exclude_ports"),
        )
        method_options.add_argument(
            "-u",
            "--usernames",
            action="store",
            dest="usernames",
            default=Config.settings.usernames,
            help=_("username_list"),
        )
        method_options.add_argument(
            "-U",
            "--users-list",
            action="store",
            dest="usernames_list",
            default=Config.settings.usernames_list,
            help=_("username_from_file"),
        )
        method_options.add_argument(
            "-p",
            "--passwords",
            action="store",
            dest="passwords",
            default=Config.settings.passwords,
            help=_("password_separator"),
        )
        method_options.add_argument(
            "-P",
            "--passwords-list",
            action="store",
            dest="passwords_list",
            default=Config.settings.passwords_list,
            help=_("read_passwords"),
        )
        method_options.add_argument(
            "-g",
            "--ports",
            action="store",
            dest="ports",
            default=Config.settings.ports,
            help=_("port_separator"),
        )
        method_options.add_argument(
            "--schema",
            action="store",
            dest="schema",
            default=Config.settings.schema,
            help=_("schema_selector"),
        )
        method_options.add_argument(
            "--user-agent",
            action="store",
            dest="user_agent",
            default=Config.settings.user_agent,
            help=_("select_user_agent"),
        )
        method_options.add_argument(
            "-T",
            "--timeout",
            action="store",
            dest="timeout",
            default=Config.settings.timeout,
            type=float,
            help=_("timeout"),
        )
        method_options.add_argument(
            "-w",
            "--time-sleep-between-requests",
            action="store",
            dest="time_sleep_between_requests",
            default=Config.settings.time_sleep_between_requests,
            type=float,
            help=_("time_to_sleep"),
        )
        method_options.add_argument(
            "-r",
            "--range",
            action="store_true",
            default=Config.settings.scan_ip_range,
            dest="scan_ip_range",
            help=_("range"),
        )
        method_options.add_argument(
            "-s",
            "--sub-domains",
            action="store_true",
            default=Config.settings.scan_subdomains,
            dest="scan_subdomains",
            help=_("subdomains"),
        )
        method_options.add_argument(
            "-d",
            "--skip-service-discovery",
            action="store_true",
            default=Config.settings.skip_service_discovery,
            dest="skip_service_discovery",
            help=_("skip_service_discovery"),
        )
        method_options.add_argument(
            "-t",
            "--thread-per-host",
            action="store",
            default=Config.settings.thread_per_host,
            type=int,
            dest="thread_per_host",
            help=_("thread_number_connections"),
        )
        method_options.add_argument(
            "-M",
            "--parallel-module-scan",
            action="store",
            default=Config.settings.parallel_module_scan,
            type=int,
            dest="parallel_module_scan",
            help=_("thread_number_modules"),
        )
        method_options.add_argument(
            "--set-hardware-usage",
            action="store",
            dest="set_hardware_usage",
            default=Config.settings.set_hardware_usage,
            help=_("set_hardware_usage"),
        )
        method_options.add_argument(
            "-R",
            "--socks-proxy",
            action="store",
            dest="socks_proxy",
            default=Config.settings.socks_proxy,
            help=_("outgoing_proxy"),
        )
        method_options.add_argument(
            "--retries",
            action="store",
            dest="retries",
            type=int,
            default=Config.settings.retries,
            help=_("connection_retries"),
        )
        method_options.add_argument(
            "--ping-before-scan",
            action="store_true",
            dest="ping_before_scan",
            default=Config.settings.ping_before_scan,
            help=_("ping_before_scan"),
        )
        method_options.add_argument(
            "-K",
            "--scan-compare",
            action="store",
            dest="scan_compare_id",
            default=Config.settings.scan_compare_id,
            help=_("compare_scans"),
        )
        method_options.add_argument(
            "-J",
            "--compare-report-path",
            action="store",
            dest="compare_report_path_filename",
            default=Config.settings.compare_report_path_filename,
            help=_("compare_report_path_filename"),
        )
        method_options.add_argument(
            "-W",
            "--wordlist",
            action="store",
            default=Config.settings.read_from_file,
            dest="read_from_file",
            help=_("user_wordlist"),
        )
        method_options.add_argument(
            "-H",
            "--add-http-header",
            action="append",
            default=Config.settings.http_header,
            dest="http_header",
            help=_("http_header"),
        )

        # API Options
        api_options = self.add_argument_group(_("API"), _("API_options"))
        api_options.add_argument(
            "--start-api",
            action="store_true",
            dest="start_api_server",
            default=Config.api.start_api_server,
            help=_("start_api_server"),
        )
        api_options.add_argument(
            "--api-host",
            action="store",
            dest="api_hostname",
            default=Config.api.api_hostname,
            help=_("API_host"),
        )
        api_options.add_argument(
            "--api-port",
            action="store",
            dest="api_port",
            default=Config.api.api_port,
            help=_("API_port"),
        )
        api_options.add_argument(
            "--api-debug-mode",
            action="store_true",
            dest="api_debug_mode",
            default=Config.api.api_debug_mode,
            help=_("API_debug"),
        )
        api_options.add_argument(
            "--api-access-key",
            action="store",
            dest="api_access_key",
            default=Config.api.api_access_key,
            help=_("API_access_key"),
        )
        api_options.add_argument(
            "--api-client-whitelisted-ips",
            action="store",
            dest="api_client_whitelisted_ips",
            default=Config.api.api_client_whitelisted_ips,
            help=_("define_white_list"),
        )
        api_options.add_argument(
            "--api-access-log",
            action="store",
            dest="api_access_log",
            default=Config.api.api_access_log,
            help=_("API_access_log_file"),
        )
        api_options.add_argument(
            "--api-cert",
            action="store",
            dest="api_cert",
            help=_("API_cert"),
        )
        api_options.add_argument(
            "--api-cert-key",
            action="store",
            dest="api_cert_key",
            help=_("API_cert_key"),
        )

    def parse_arguments(self):
        """
        check all rules and requirements for ARGS

        Args:
            api_forms: values from nettacker.api

        Returns:
            all ARGS with applied rules
        """
        # Checking Requirements
        options = self.api_arguments or self.parse_args()

        if options.language not in self.languages:
            die_failure("Please select one of these languages {0}".format(self.languages))

        # Check Help Menu
        if options.show_help_menu:
            self.print_help()
            log.write("\n\n")
            log.write(_("license"))
            die_success()

        # Check version
        if options.show_version:
            log.info(
                _("current_version").format(
                    TerminalCodes.YELLOW.value,
                    version_info()[0],
                    TerminalCodes.RESET.value,
                    TerminalCodes.CYAN.value,
                    version_info()[1],
                    TerminalCodes.RESET.value,
                    TerminalCodes.GREEN.value,
                )
            )
            die_success()

        if options.show_all_modules:
            log.info(_("loading_modules"))
            for module in self.modules:
                log.info(
                    _("module_profile_full_information").format(
                        TerminalCodes.CYAN.value,
                        module,
                        TerminalCodes.GREEN.value,
                        ", ".join(
                            [
                                "{key}: {value}".format(key=key, value=self.modules[module][key])
                                for key in self.modules[module]
                            ]
                        ),
                    )
                )
            die_success()

        if options.show_all_profiles:
            log.info(_("loading_profiles"))
            for profile in self.profiles:
                log.info(
                    _("module_profile_full_information").format(
                        TerminalCodes.CYAN.value,
                        profile,
                        TerminalCodes.GREEN.value,
                        ", ".join(self.profiles[profile]),
                    )
                )
            die_success()

        # API mode
        if options.start_api_server:
            if "--start-api" in sys.argv and self.api_arguments:
                die_failure(_("cannot_run_api_server"))
            from nettacker.api.engine import start_api_server

            if options.api_client_whitelisted_ips:
                if isinstance(options.api_client_whitelisted_ips, str):
                    options.api_client_whitelisted_ips = options.api_client_whitelisted_ips.split(
                        ","
                    )
                    whitelisted_ips = []
                    for ip in options.api_client_whitelisted_ips:
                        if is_single_ipv4(ip) or is_single_ipv6(ip):
                            whitelisted_ips.append(ip)
                        elif (
                            is_ipv4_range(ip)
                            or is_ipv6_range(ip)
                            or is_ipv4_cidr(ip)
                            or is_ipv6_cidr(ip)
                        ):
                            whitelisted_ips += generate_ip_range(ip)
                    options.api_client_whitelisted_ips = whitelisted_ips
            start_api_server(options)

        # Check the target(s)
        if not (options.targets or options.targets_list) or (
            options.targets and options.targets_list
        ):
            # self.print_help()
            # write("\n")
            die_failure(_("error_target"))
        if options.targets:
            options.targets = list(set(options.targets.split(",")))
        if options.targets_list:
            try:
                options.targets = list(
                    set(open(options.targets_list, "rb").read().decode().split())
                )
            except Exception:
                die_failure(_("error_target_file").format(options.targets_list))

        # check for modules
        if not (options.selected_modules or options.profiles):
            die_failure(_("scan_method_select"))
        if options.selected_modules:
            if options.selected_modules == "all":
                options.selected_modules = list(set(self.modules.keys()))
                options.selected_modules.remove("all")
            else:
                options.selected_modules = list(set(options.selected_modules.split(",")))
            for module_name in options.selected_modules:
                if module_name not in self.modules:
                    die_failure(_("scan_module_not_found").format(module_name))
        if options.profiles:
            if not options.selected_modules:
                options.selected_modules = []
            if options.profiles == "all":
                options.selected_modules = list(set(self.modules.keys()))
                options.selected_modules.remove("all")
            else:
                options.profiles = list(set(options.profiles.split(",")))
                for profile in options.profiles:
                    if profile not in self.profiles:
                        die_failure(_("profile_404").format(profile))
                    for module_name in self.profiles[profile]:
                        if module_name not in options.selected_modules:
                            options.selected_modules.append(module_name)
        # threading & processing
        if options.set_hardware_usage not in {"low", "normal", "high", "maximum"}:
            die_failure(_("wrong_hardware_usage"))
        options.set_hardware_usage = common_utils.select_maximum_cpu_core(
            options.set_hardware_usage
        )

        options.thread_per_host = int(options.thread_per_host)
        if options.thread_per_host < 1:
            options.thread_per_host = 1
        options.parallel_module_scan = int(options.parallel_module_scan)
        if options.parallel_module_scan < 1:
            options.parallel_module_scan = 1

        # Check for excluding modules
        if options.excluded_modules:
            options.excluded_modules = options.excluded_modules.split(",")
            if "all" in options.excluded_modules:
                die_failure(_("error_exclude_all"))
            for excluded_module in options.excluded_modules:
                if excluded_module in options.selected_modules:
                    options.selected_modules.remove(excluded_module)
        # Check port(s)
        if options.ports:
            tmp_ports = set()
            for port in options.ports.split(","):
                try:
                    if "-" in port:
                        for port_number in range(
                            int(port.split("-")[0]), int(port.split("-")[1]) + 1
                        ):
                            tmp_ports.add(port_number)
                    else:
                        tmp_ports.add(int(port))
                except Exception:
                    die_failure(_("ports_int"))
            options.ports = list(tmp_ports)
        # Check schema(s)
        if options.schema:
            tmp_schema = {schema.strip().lower() for schema in options.schema.split(",")}
            allowed_schema = {"http", "https"}

            if tmp_schema - allowed_schema:
                die_failure(_("invalid_schema"))
            else:
                options.schema = list(tmp_schema)
        # Check for excluded ports
        if options.excluded_ports:
            tmp_excluded_ports = set()

            for excluded_port in options.excluded_ports.split(","):
                try:
                    if "-" in excluded_port:
                        for excluded_port_number in range(
                            int(excluded_port.split("-")[0]), int(excluded_port.split("-")[1]) + 1
                        ):
                            tmp_excluded_ports.add(excluded_port_number)
                    else:
                        tmp_excluded_ports.add(int(excluded_port))
                except Exception:
                    die_failure(_("ports_int"))
            options.excluded_ports = list(tmp_excluded_ports)

        if options.user_agent == "random_user_agent":
            options.user_agents = open(Config.path.user_agents_file).read().split("\n")

        # Check user list
        if options.usernames:
            options.usernames = list(set(options.usernames.split(",")))
        elif options.usernames_list:
            try:
                options.usernames = list(set(open(options.usernames_list).read().split("\n")))
            except Exception:
                die_failure(_("error_username").format(options.usernames_list))
        # Check password list
        if options.passwords:
            options.passwords = list(set(options.passwords.split(",")))
        elif options.passwords_list:
            try:
                options.passwords = list(set(open(options.passwords_list).read().split("\n")))
            except Exception:
                die_failure(_("error_passwords").format(options.passwords_list))
        # Check custom wordlist
        if options.read_from_file:
            try:
                open(options.read_from_file).read().split("\n")
            except Exception:
                die_failure(_("error_wordlist").format(options.read_from_file))
        # Check output file
        try:
            temp_file = open(options.report_path_filename, "w")
            temp_file.close()
        except Exception:
            die_failure(_("file_write_error").format(options.report_path_filename))
        # Check Graph
        if options.graph_name:
            if options.graph_name not in self.graphs:
                die_failure(_("graph_module_404").format(options.graph_name))
            if not (
                options.report_path_filename.endswith(".html")
                or options.report_path_filename.endswith(".htm")
            ):
                log.warn(_("graph_output"))
                options.graph_name = None
        # check modules extra args
        if options.modules_extra_args:
            all_args = {}
            for args in options.modules_extra_args.split("&"):
                value = args.split("=")[1]
                if value.lower() == "true":
                    value = True
                elif value.lower() == "false":
                    value = False
                elif "." in value:
                    try:
                        value = float(value)
                    except Exception:
                        pass
                elif "{" in value or "[" in value:
                    try:
                        value = json.loads(value)
                    except Exception:
                        pass
                else:
                    try:
                        value = int(value)
                    except Exception:
                        pass
                all_args[args.split("=")[0]] = value
            options.modules_extra_args = all_args

        options.timeout = float(options.timeout)
        options.time_sleep_between_requests = float(options.time_sleep_between_requests)
        options.retries = int(options.retries)

        self.arguments = options



from sqlalchemy import create_engine, text
from sqlalchemy.exc import OperationalError

from nettacker.config import Config
from nettacker.database.models import Base


def postgres_create_database():
    """
    when using postgres database, this is the function that is used to
    create the database for the first time when you the nettacker run module.
    """
    try:
        engine = create_engine(
            "postgresql+psycopg2://{username}:{password}@{host}:{port}/{name}".format(
                **Config.db.as_dict()
            )
        )
        Base.metadata.create_all(engine)
    except OperationalError:
        # if the database does not exist
        engine = create_engine(
            "postgresql+psycopg2://{username}:{password}@{host}:{port}/postgres".format(
                **Config.db.as_dict()
            )
        )
        conn = engine.connect()
        conn = conn.execution_options(isolation_level="AUTOCOMMIT")
        conn.execute(text(f"CREATE DATABASE {Config.db.name}"))
        conn.close()
        engine = create_engine(
            "postgresql+psycopg2://{username}:{password}@{host}:{port}/{name}".format(
                **Config.db.as_dict()
            )
        )
        Base.metadata.create_all(engine)



import json

from nettacker.config import Config


def build_report(compare_result):
    """
    generate a report based on result of comparision b/w scans

    Args:
        compare_result: dict with result of the compare

    Returns:
        Compare report in HTML
    """
    data = (
        open(Config.path.web_static_dir / "report/compare_report.html")
        .read()
        .replace("__data_will_locate_here__", json.dumps(compare_result))
    )
    return data



import telnetlib

from nettacker.core.lib.base import BaseEngine, BaseLibrary


class TelnetLibrary(BaseLibrary):
    client = telnetlib.Telnet

    def brute_force(self, host, port, username, password, timeout):
        connection = self.client(host, port, timeout)
        connection.read_until(b"login: ")
        connection.write(username.encode("utf-8") + b"\n")
        connection.read_until(b"Password: ")
        connection.write(password.encode("utf-8") + b"\n")
        connection.close()

        return {
            "host": host,
            "port": port,
            "username": username,
            "password": password,
        }


class TelnetEngine(BaseEngine):
    library = TelnetLibrary
